#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

import os
import re
import random
import pickle
import torch
import lmdb
import numpy as np
from ase.io import read
from ase import Atoms, Atom
from ase.db import connect
from torch_geometric.data import Data
from ase.geometry import get_distances
from mofstructure import mofdeconstructor
# from orb_models.forcefield import atomic_system
from fairmofsyncondition.read_write import filetyper


def read_and_return_ase_atoms(filename):
    """
    Function to read the ase atoms

    **parameter**
        filename: string
    """
    ase_atoms = read(filename)
    return ase_atoms


def write_ase_atoms(ase_atoms, filename):
    """
    Function to write the ase atoms

    **parameter**
        ase_atoms: ase.Atoms object
        filename: string
    """
    ase_atoms.write(filename)


def ase_coordinate(filename):
    """
    Read any ASE readable file and returns coordinates and lattices
    which should be use for setting up AMS calculations.

    **parameter**
        filename  (string) : Name of file containing the coordinate

    **Returns**
        ase_coord (list) : List of coordinate strings
        lattice (list) : List of lattice vectors strings
    """
    molecule = read(filename)
    atoms = Atoms(molecule)
    ase_cell = atoms.get_cell(complete=True)
    elements = atoms.get_chemical_symbols()
    positions = atoms.get_positions()
    ase_coord = []
    for ele, xyz in zip(elements, positions):
        cods = '\t'.join([ele]+[str(i) for i in xyz])
        ase_coord.append(cods)
    lattice = []
    for i in range(3):
        a = [' '] + [str(i) for i in ase_cell[i]]
        b = '\t'.join(a)
        lattice.append(b)
    return ase_coord, lattice


def gjf_coordinate(filename):
    """
    Reading coordinates from a gaussian .gjf file

    **parameter**
        filename  (string) : Name of file containing the coordinate

    **Returns**
        coords : List of coordinate strings
        lattice (list) : List of lattice vectors strings
    """
    qc_input = filetyper.get_contents(filename)
    file_lines = []
    for line in qc_input:
        file_lines.append(line.split())

    coords = []
    lattice = []
    ase_line = file_lines[2]
    if 'ASE' not in ase_line:
        for row in file_lines[6:]:
            if len(row) > 0:
                if 'Tv' not in row:
                    b = '\t'.join(row)
                    coords.append(b)
                else:
                    b = '\t'.join(row[1:])
                    lattice.append(b)
            else:
                break
    else:
        for row in file_lines[5:]:
            if len(row) > 0:
                if 'TV' not in row:
                    b = '\t'.join(row)
                    coords.append(b)
                else:
                    b = '\t'.join(row[1:])
                    lattice.append(b)
            else:
                break

    return coords, lattice


def xyz_coordinates(filename):
    """
    Read any xyz coordinate file

    **parameter**
        filename  (string) : Name of file containing the coordinate

    **Returns**
        coords : List of coordinate strings
    """
    qc_input = filetyper.get_contents(filename)
    coords = []
    file_lines = []
    for line in qc_input:
        file_lines.append(line.split())
    for row in file_lines[2:]:
        a = [' '] + row
        b = '\t'.join(a)
        coords.append(b)
    return coords


def check_periodicity(filename):
    """
    Function to check periodicity in an scm output file

    **parameter**
        filename  (string) : Name of file containing the coordinate
    """
    qc_input = filetyper.get_contents(filename)
    verdict = False
    for line in qc_input:
        if 'Lattice vectors (angstrom)' in line:
            verdict = True
            break
    return verdict


def scm_out(qcin):
    """
    Extract coordinates from scm output files

    **parameter**
        qcin  (string) : scm output file

    **return**
        coords (list) : list of coordinates
        lattice_coords (list) : list of lattice coordinates
    """
    qc_input = filetyper.get_contents(qcin)
    verdict = check_periodicity(qcin)
    coords = []
    lattice_coords = []
    lattice = []
    length_value = []
    if verdict:
        cods = filetyper.get_section(
            qc_input,
            'Index Symbol   x (angstrom)   y (angstrom)   z (angstrom)',
            'Lattice vectors (angstrom)',
            1,
            -2
            )

        for lines in cods:
            data = lines.split()
            length_value.append(data[0])
            b = '\t'.join(data[1:])
            coords.append(b)
        lat_index = 0
        for i, line in enumerate(qc_input):
            data = line.split()
            lattice.append(data)
            if 'Lattice vectors (angstrom)' in line:
                lat_index = i

        parameters = [lattice[lat_index+1],
                      lattice[lat_index+2],
                      lattice[lat_index+3]
                      ]

        for line in parameters:
            a = line[1:]
            if len(a) > 2:
                b = '\t'.join(a)
                lattice_coords.append(b)

    else:
        cods = filetyper.get_section(
            qc_input,
            'Index Symbol   x (angstrom)   y (angstrom)   z (angstrom)',
            'Total System Charge', 1, -2
            )
        for lines in cods:
            data = lines.split()
            length_value.append(data[0])
            b = '\t'.join(data[1:])
            coords.append(b)
        # length = str(len(length_value))
        lattice_coords = ['']
    return coords, lattice_coords


def qchemcout(filename):
    """
    Read coordinates from qchem output file

    **parameter**
        filename  (string) : Name of file containing the coordinate

    **Returns**
        coords : List of coordinate strings
    """
    qc_input = filetyper.get_contents(filename)
    cods = filetyper.get_section(qc_input,
                                 'OPTIMIZATION CONVERGED',
                                 'Z-matrix Print:',
                                 5,
                                 -2
                                 )
    # cods = filetyper.get_section(qc_input, '$molecule', '$end', 2, -1)
    coords = []
    for row in cods:
        data = row.split()
        b = '\t'.join(data[1:])
        coords.append(b)
    return coords


def qchemin(filename):
    """
    Read coordinates from qchem input file

    **parameter**
        filename (string) : filename

    **Returns**
        coords : list of coordinate strings
    """
    qc_input = filetyper.get_contents(filename)
    coords = filetyper.get_section(qc_input, '$molecule', '$end', 2, -1)
    return coords


def format_coords(coords, atom_labels):
    """
    create coords containing symbols and positions

    **parameters**
        coords (list) : list of coordinates
        atom_labels (list) : list of atom labels

    **returns**
        coordinates (list) : list of formatted coordinates
    """

    coordinates = []
    # file_obj.write('%d\n\n' %len(atom_types))
    for labels, row in zip(atom_labels, coords):
        b = [labels] + [str(atom)+' ' for atom in row]
        printable_row = '\t'.join(b)
        coordinates.append(printable_row + '\n')
    return coordinates


def coordinate_definition(filename):
    """
    define how coordinates should be extracted
    """
    # print (filename)
    # Robust algorithm for finding file extention (check)
    iter_index = re.finditer(r'\.', filename)
    check = [filename[i.span()[0]+1:] for i in iter_index][-1]
    coords, lattice = [], []
    # check = filename.split('.')[1]
    if check == 'gjf':
        coords, lattice = gjf_coordinate(filename)
    elif check == 'xyz':
        coords = xyz_coordinates(filename)
    elif check == 'out':
        coords, lattice = scm_out(filename)
    elif check == 'cout':
        coords = qchemcout(filename)
    elif check == 'cin':
        coords = qchemin(filename)
    else:
        coords, lattice = ase_coordinate(filename)

    return coords, lattice


def collect_coords(filename):
    '''
    Collect coordinates

    **parameters**
        filename (string) : filename

    **returns**
        elements (list) : list of elements
        positions (numpy array) : numpy array of positions
        cell (numpy array) : numpy array of
        cell parameters if present in the file
    '''
    coords, lattice = coordinate_definition(filename)
    elements = []
    positions = []
    cell = []
    for lines in coords:
        data = lines.split()
        elements.append(data[0])
        positions.append([float(i) for i in data[1:]])

    positions = np.array(positions)

    if len(lattice) != 0:
        cell = np.array([[float(i) for i in j.split()] for j in lattice])

    return elements, positions, cell


def load_data_as_ase(filename):
    """
    Load data as an ase atoms object
    **parameter**
        filename (string) : Any file type that has been defined in this module
                            including ase readable filetypes
    **return**
        ase_atoms : ase atoms object
    """
    elements, positions, cell = collect_coords(filename)
    ase_atoms = Atoms(symbols=elements, positions=positions)
    if len(cell) > 0:
        ase_atoms = Atoms(symbols=elements,
                          positions=positions,
                          cell=cell,
                          pbc=True
                          )
    return ase_atoms


def  charge_and_ase_from_ams_gfn(filename):
    """
    Extract charge and ase atoms from an AMS gfn output file

    **parameter**
        filename (string) : AMS gfn output file

    **return**
        ase_atoms (ase Atoms object): ase atoms object
        charge (float): charge of the system
    """
    contents = filetyper.get_contents(filename)
    verdict = check_periodicity(filename)
    positions = []
    atoms_symbols = []
    lattice = []
    charge = []
    charge_in_file = filetyper.get_section(
            contents, 'Index   Atom          Charge      Population', 'Mulliken Shell Charges', 1, -4
            )
    for lines in charge_in_file:
        charge.append(float(lines.split()[2]))

    if verdict:
        cods = filetyper.get_section(
            contents,
            'Index Symbol   x (angstrom)   y (angstrom)   z (angstrom)',
            'Lattice vectors (angstrom)',
            1,
            -2
            )

        for lines in cods:
            data = lines.split()
            atoms_symbols.append(data[1])
            positions.append([float(i) for i in data[2:]])

        file_lattice = filetyper.get_section(
            contents,
            'Lattice vectors (angstrom)', 'Unit cell volume', 1,-2)
        for lines in file_lattice:
            data = lines.split()
            lattice.append([float(i) for i in data[1:]])
        ase_atoms = Atoms(symbols=atoms_symbols,
                          positions=positions,
                          cell=lattice,
                          pbc=True
                          )
    else:
        cods = filetyper.get_section(
            contents,
            'Index Symbol   x (angstrom)   y (angstrom)   z (angstrom)',
            'Total System Charge', 1, -2
            )
        for lines in cods:
            data = lines.split()
            atoms_symbols.append(data[1])
            positions.append([float(i) for i in data[2:]])
        ase_atoms = Atoms(symbols=atoms_symbols, positions=positions)

    return ase_atoms, charge


def compute_esp(atoms: Atoms, charges: np.ndarray, eps: float = 1e-12) -> np.ndarray:
    """
    Compute the electrostatic potential (ESP) at each atomic position due to all other atoms,
    using the point charge approximation and the minimum image convention under periodic boundary conditions (PBC).

    The electrostatic potential at atom *i* is computed as:

        V_i = Σ_{j ≠ i} (q_j / r_ij)

    where:
        - V_i is the electrostatic potential at atom *i*,
        - q_j is the charge of atom *j*,
        - r_ij is the shortest distance between atoms *i* and *j*, accounting for PBC.

    **Parameters**
        - atoms : ase.Atoms
            An ASE Atoms object representing the system. The simulation cell and periodic boundary conditions (PBC) must be defined.
        - charges : np.ndarray
        A 1D NumPy array of atomic point charges (in units of elementary charge, e), with length equal to the number of atoms.
        - eps : float, optional
        A small constant added to distances to prevent division by zero (default is 1e-12 Å).

    **Returns**
        -np.ndarray
            A 1D NumPy array containing the electrostatic potential at each atomic position (in units of e/Å).

    **Notes**
        - To convert the ESP values to electronvolts (eV), multiply the result by 14.3996
        (since 1 e / (4 * π * ε₀ * Å) ≈ 14.3996 eV·Å/e).
        - Self-interactions are excluded by setting the diagonal of the distance matrix to ∞.
        - Distance calculations use ASE's `get_distances` function, which applies the minimum image convention under PBC.
    """
    if not isinstance(charges, np.ndarray):
        charges = np.array(charges, dtype=float)
    distances = atoms.get_all_distances(mic=atoms.pbc.any())
    np.fill_diagonal(distances, np.inf)
    esp = np.sum(charges / (distances + eps), axis=1)
    return esp


# def ase_graph(input_system):
#     """
#     Create a graph from an ase atoms object

#     **parameter**
#         **input_system** : Atoms or Atom object or meolcular file name e.g molecule.xyz or mof.cif

#     **return**
#         graph object: ase graph object
#     """
#     if isinstance(input_system, Atoms) or isinstance(input_system, Atom):
#         graph = atomic_system.ase_atoms_to_atom_graphs(input_system)
#     else:
#         ase_atoms = load_data_as_ase(input_system)
#         graph = atomic_system.ase_atoms_to_atom_graphs(ase_atoms)
#     return graph


def xtb_input(filename):
    """
    Creating a gfn-xtb input file from any ase readable filetype or filetype
    that can be read by this module.

    **parameter**
        filename (string) : Any file type that has been defined in this module

    **return**
        xtb_coords : list of strings containing xtb input
    """
    elements, positions, cell = collect_coords(filename)
    xtb_coords = []
    # xtb_coords.append('> cat coord \n')
    xtb_coords.append('$coord angs\n')
    for labels, row in zip(elements, positions):
        tmp_coord = [str(atom) + ' ' for atom in row] + [' '] + [labels]
        xtb_coords.append('\t'.join(tmp_coord) + '\n')
    if len(cell) > 0:
        xtb_coords.append('$periodic ' + str(len(cell)) + '\n')
        xtb_coords.append('$lattice angs \n')
        for lattice in cell:
            lat_vector = '\t'.join(lattice) + '\n'
            xtb_coords.append(lat_vector)
    xtb_coords.append('$end')
    # xtb_coords.append('> xtb coord\n')
    return xtb_coords


def ase_to_xtb(ase_atoms):
    """
    Create a gfn-xtb input from an ase atom object.

    **parameter**
        ase_atoms (ase Atoms or Atom): The ase atoms object to be converted.

    **return**
        xtb_coords = ase_to_xtb_coords(ase_atoms)
    """
    check_pbc = ase_atoms.get_pbc()
    ase_cell = []
    xtb_coords = []
    if any(check_pbc):
        ase_cell = ase_atoms.get_cell(complete=True)
    elements = ase_atoms.get_chemical_symbols()
    positions = ase_atoms.get_positions()
    # xtb_coords.append('> cat coord \n')
    xtb_coords.append('$coord angs\n')
    for labels, row in zip(elements, positions):
        tmp_coord = [str(atom) + ' ' for atom in row] + [' '] + [labels]
        xtb_coords.append('\t'.join(tmp_coord) + '\n')
    if len(ase_cell) > 0:
        xtb_coords.append('$periodic cell vectors \n')
        # xtb_coords.append('$lattice angs \n')
        for lattice in ase_cell:
            tmp_lattice = [str(lat) + ' ' for lat in lattice] + [' ']
            xtb_coords.append('\t'.join(tmp_lattice) + '\n')
    xtb_coords.append('$end')
    return xtb_coords


def get_pairwise_connections(graph):
    """
    Extract unique pairwise connections from an
    adjacency dictionary efficiently.

    **Parameters**
        graph (dict):
            An adjacency dictionary where keys are nodes
            and values are arrays or lists of nodes
            representing neighbors.

    **returns**
        list of tuple
            A list of unique pairwise connections,
            each represented as a tuple (i, j) where i < j.

    """
    pairwise_connections = []
    seen = set()

    for node, neighbors in graph.items():
        for neighbor in neighbors:
            edge = (min(node, neighbor), max(node, neighbor))
            if edge not in seen:
                seen.add(edge)
                pairwise_connections.append(edge)
    return pairwise_connections


def calculate_distances(pair_indices, ase_atoms, mic=True):
    """
    Calculate distances between pairs of atoms in an ase atoms object.
    """
    return np.array([
        ase_atoms.get_distance(pair[0], pair[1], mic=mic)
        for pair in pair_indices])


def ase_to_pytorch_geometric(input_system):
    """
    Convert an ASE Atoms object to a PyTorch Geometric graph

    **parameters**
        input_system (ASE.Atoms or ASE.Atom or filename):
        The input system to be converted.

    **returns**
        torch_geometric.data.Data: The converted PyTorch Geometric Data object.
    """

    if isinstance(input_system, Atoms) or isinstance(input_system, Atom):
        ase_atoms = input_system
    else:
        ase_atoms = load_data_as_ase(input_system)
    mic = ase_atoms.pbc.any()
    if mic:
        lattice_parameters = torch.tensor(np.array(ase_atoms.cell),
                                          dtype=torch.float
                                          )
    else:
        lattice_parameters = torch.tensor(np.zeros((3, 3)),
                                          dtype=torch.float
                                          )

    graph, _ = mofdeconstructor.compute_ase_neighbour(ase_atoms)
    pair_connection = np.array(get_pairwise_connections(graph))
    distances = calculate_distances(pair_connection, ase_atoms, mic)
    nodes = np.array([[atom.number, *atom.position] for atom in ase_atoms])

    node_features = torch.tensor(nodes, dtype=torch.float)
    edge_index = torch.tensor(pair_connection,
                              dtype=torch.long).t().contiguous()
    edge_attr = torch.tensor(distances,
                             dtype=torch.float).unsqueeze(1)
    data = Data(x=node_features,
                edge_index=edge_index,
                edge_attr=edge_attr,
                lattice=lattice_parameters)
    return data


def pytorch_geometric_to_ase(data):
    """
    Convert a PyTorch Geometric Data object back to an ASE Atoms object.

    **Parameters**
        data (torch_geometric.data.Data): The PyTorch Geometric Data object.

    **Returns**
        ase_atoms (ase.Atoms): The converted ASE Atoms object.
    """
    node_features = data.x.numpy() if isinstance(data.x,
                                                 torch.Tensor) else data.x
    atomic_numbers = node_features[:, 0].astype(int)
    positions = node_features[:, 1:4]

    lattice = data.lattice.numpy() if isinstance(data.lattice,
                                                 torch.Tensor
                                                 ) else data.lattice

    ase_atoms = Atoms(
        numbers=atomic_numbers,
        positions=positions,
        cell=lattice,
        pbc=(lattice.any())
    )

    return ase_atoms


def prepare_dataset(ase_obj, energy):
    """
    Prepares a dataset from ASE Atoms objects and their
    corresponding energy values.

    **parameters**
        ase_obj (ASE.Atoms): ASE Atoms object.
        energy (float): Energy value of the crystal structure.

    **returns**
        torch_geometric.data.Data: PyTorch Geometric Data object
        with input features, edge indices, and energy value.
    """
    data = ase_to_pytorch_geometric(ase_obj)
    data.y = torch.tensor([energy], dtype=torch.float)
    return data


def data_from_aseDb(path_to_db, num_data=25000):
    """
    Load data from ASE database and prepare it for training.

    **parameters**
        path_to_db (str): Path to the ASE database file.

    **returns**
        list: List of PyTorch Geometric Data objects for training.
    """
    dataset = []
    counter = 0
    db = connect(path_to_db)
    for row in db.select():
        data = prepare_dataset(row.toatoms(), row.r_energy)
        dataset.append(data)
        if counter >= num_data:
            break
        counter += 1
    return dataset


def ase_database_to_lmdb(ase_database, lmdb_path):
    """
    Converts an ASE database into an LMDB file for
    efficient storage and retrieval.

    **parameter**
        ase_database (str): path to ase database.
        lmdb_path (str): Path to the LMDB file where
            the dataset will be saved.
    """
    os.makedirs(os.path.dirname(lmdb_path), exist_ok=True)

    try:
        with connect(ase_database) as db:
            with lmdb.open(lmdb_path, map_size=int(1e12)) as lmdb_env:
                with lmdb_env.begin(write=True) as txn:
                    count = 0
                    for i, row in enumerate(db.select()):
                        data = prepare_dataset(row.toatoms(), row.r_energy)
                        txn.put(f"{i}".encode(), pickle.dumps(data))
                        count += 1
                    txn.put(b"__len__", pickle.dumps(count))
        print(f"Data successfully saved to {lmdb_path} with {count} entries.")
    except lmdb.Error as e:
        print(f"An error occurred with LMDB: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

class LMDBDataset:
    """
    A class for loading PyTorch data stored in an LMDB
    file. The code is originally
    intended for graph structured graphs that can work with
    pytorch_geometric data.
    But it should also load all types of PyTorch data.

    This class enables on-the-fly loading of serialized data
    stored in LMDB format, providing an efficient way to handle
    large datasets that cannot fit into memory.

    **parameters**
        lmdb_path (str): Path to the LMDB file containing the dataset.

    **Attributes**
        lmdb_env (lmdb.Environment): The LMDB environment for data access.
        length (int): The total number of entries in the dataset.

    **Methods**
        __len__(): Returns the total number of samples in the dataset.
        __getitem__(idx): Retrieves the sample at the specified index.
        split_data(train_size, random_seed, shuffle): Lazily
        returns train and test data.

    **Examples**

    This class provides an efficient way for loading huge datasets without
    consuming so much memory.

        data = coords_library.LMDBDataset(path_to_lmdb)

        Length of the dataset

        print(len(data))

        # Accessing a sample at index 0

        sample = data[0]

        print(sample.x.shape)

        print(sample)

        # Accessing a list of samples at different indexes

        samples = data[[1,4,8,9,18, 50]]

    """

    def __init__(self, lmdb_path):
        try:
            self.lmdb_env = lmdb.open(lmdb_path, readonly=True, lock=False)
            with self.lmdb_env.begin() as txn:
                length_data = txn.get(b"__len__")
                if length_data is None:
                    raise ValueError(
                        f"""
                        The LMDB file at '{lmdb_path}'
                        does not contain the key '__len__'.
                        """
                        f"""Ensure the data was saved
                        correctly and includes this key.
                        """
                    )
                self.length = pickle.loads(length_data)
        except ValueError as ve:
            raise RuntimeError(
                f"""
                ValueError: {ve}\nCheck if the
                LMDB file is correctly created with a
                '__len__' key."""
            ) from ve
        except lmdb.Error as le:
            raise RuntimeError(
                f"""
                An LMDB error occurred while
                accessing the file at '{lmdb_path}': {le}
                """
            ) from le
        except Exception as e:
            raise RuntimeError(
                f"""
                An unexpected error occurred
                while initializing the dataset: {e}
                """
            ) from e

    def __len__(self):
        """
        Returns the total number of samples in the dataset.

        Returns
            int: The number of samples in the dataset.
        """
        return self.length

    def __getitem__(self, idx):
        """
        Retrieves the sample(s) at the specified index or
        indices from the LMDB file.

        **parameters**
            idx (int or list of int): The index or indices of
            the sample(s) to retrieve.

        **Returns**
            Any or list: The deserialized data corresponding
            to the specified index/indices.
        """
        if isinstance(idx, int):
            if idx < 0 or idx >= self.length:
                raise IndexError(
                    f"""
                    Index {idx} is out of range
                    for dataset of size {self.length}.
                    """
                )
            with self.lmdb_env.begin() as txn:
                data = txn.get(f"{idx}".encode())
                if data is None:
                    raise ValueError(
                        f"""
                        No data found for index {idx}.
                        Ensure the dataset is correctly saved.
                        """
                    )
                return pickle.loads(data)
        elif isinstance(idx,
                        list) or isinstance(idx,
                                            np.ndarray) or isinstance(idx,
                                                                      tuple):
            results = []
            for i in idx:
                if i < 0 or i >= self.length:
                    raise IndexError(
                        f"""
                        Index {i} is out of range for
                        dataset of size {self.length}.
                        """
                    )
                with self.lmdb_env.begin() as txn:
                    data = txn.get(f"{i}".encode())
                    if data is None:
                        raise ValueError(
                            f"""
                            No data found for index {i}.
                            Ensure the dataset is correctly saved.
                            """
                        )
                    results.append(pickle.loads(data))
            return results
        else:
            raise TypeError(
                """
                Index must be an int or list,
                or nd.array or tuple.
                """
                )

    def split_data(self, train_size=0.8, random_seed=None, shuffle=True):
        """
        Lazily splits the dataset into train and test data with
        class-like behavior.

        Args:
            train_size (float): The proportion of the data to be used
            as the training set (default is 0.8).
            random_seed (int, optional): A random seed for reproducibility
            (default is None).
            shuffle (bool): Whether to shuffle the data before splitting
            (default is True).

        Returns:
            tuple: A tuple containing train data and test data.
        """
        indices = list(range(self.length))

        if random_seed is not None:
            random.seed(random_seed)

        if shuffle:
            random.shuffle(indices)

        split_index = int(self.length * train_size)
        train_indices = indices[:split_index]
        test_indices = indices[split_index:]

        class Subset:
            def __init__(self, parent, indices):
                self.parent = parent
                self.indices = indices

            def __len__(self):
                return len(self.indices)

            def __getitem__(self, idx):
                if isinstance(idx, int):
                    return self.parent[self.indices[idx]]
                elif isinstance(idx,
                                list) or isinstance(idx,
                                                    np.ndarray) or isinstance(idx,
                                                                              tuple):
                    return [self.parent[self.indices[i]] for i in idx]
                else:
                    raise TypeError("""
                                    Index must be an int or
                                    list, or nd.array or tuple.
                                    """
                                    )

        train_data = Subset(self, train_indices)
        test_data = Subset(self, test_indices)

        return train_data, test_data


def list_train_test_split(data, train_size=0.8, random_seed=42, shuffle=True):
    """
    A function that take Splits a list into train and test
    sets based on the specified train_size.

    **parameter**
        data (list): The input list to split.
        train_size (float): The proportion of the data to be
        used as the training set (default is 0.8).
        random_seed (int, optional): A random seed for
        reproducibility (default is None).
        shuffle (bool): Whether to shuffle the data
        before splitting (default is True).

    **return**
        train_data: indices of data to be selected for training.
        test_data: indices of data to be selected for testing.
    """
    if random_seed is not None:
        random.seed(random_seed)

    if shuffle:
        data = data.copy()
        random.shuffle(data)

    split_index = int(len(data) * train_size)
    train_data = data[:split_index]
    test_data = data[split_index:]

    return train_data, test_data



import sqlite3
import random
import json
import pandas as pd
import matplotlib.pyplot as plt
from networkx.algorithms import community
import networkx as nx


class MOFNetworkAnalysis:
    """
    A class for performing social network analysis on MOF similarity data.

    Attributes:
    ----------
    db_name : str
        The name of the SQLite database containing MOF similarity data.
    similarity_threshold : float
        The threshold for filtering similarity edges between MOFs.
    G : nx.Graph
        The graph representation of MOFs and their similarity relationships.
    unique_mofs : list
        A list of communities (clusters) of MOFs.
    filtered_data : pd.DataFrame
        The filtered similarity data used to create the graph.
    """

    def __init__(self, db_name, similarity_threshold=0.5):
        """
        Initialize the MOFNetworkAnalysis class with a database and similarity threshold.

        Parameters:
        ----------
        db_name : str
            The SQLite database file containing MOF similarity data.
        similarity_threshold : float
            The minimum similarity score to consider when creating the network.
        """
        self.db_name = db_name
        self.similarity_threshold = similarity_threshold
        self.G = nx.Graph()
        self.unique_mofs = []
        self.filtered_data = None

    def load_data(self):
        """
        Load MOF similarity data from the SQLite database and filter by similarity threshold.
        """
        conn = sqlite3.connect(self.db_name)
        query = '''
        SELECT mof_name, similar_mof, similarity FROM json_data
        '''
        data = pd.read_sql_query(query, conn)
        conn.close()

        # Filter edges based on the similarity threshold
        self.filtered_data = data[data['similarity'] > self.similarity_threshold]

    def create_graph(self):
        """
        Create a graph where MOFs are nodes, and edges represent similarity relationships.
        """
        for index, row in self.filtered_data.iterrows():
            self.G.add_edge(row['mof_name'], row['similar_mof'], weight=row['similarity'])

    def compute_centrality(self):
        """
        Compute centrality metrics: degree centrality, closeness centrality, and betweenness centrality.

        Returns:
        -------
        dict
            A dictionary containing the centrality metrics for each node in the graph.
        """
        degree_centrality = nx.degree_centrality(self.G)
        closeness_centrality = nx.closeness_centrality(self.G)
        betweenness_centrality = nx.betweenness_centrality(self.G)

        return {
            'degree_centrality': degree_centrality,
            'closeness_centrality': closeness_centrality,
            'betweenness_centrality': betweenness_centrality
        }

    def detect_communities(self):
        """
        Identify communities (clusters) of MOFs using a community detection algorithm.

        Returns:
        -------
        list
            A list of communities, where each community is a list of MOFs.
        """
        communities = community.greedy_modularity_communities(self.G, weight='weight')
        self.unique_mofs = [list(community) for community in communities]
        return self.unique_mofs

    def visualize_network(self, centrality_metrics, output_image='mof_network.png'):
        """
        Visualize the MOF network with colored communities and save the figure.

        Parameters:
        ----------
        centrality_metrics : dict
            Centrality metrics for each node in the graph.
        output_image : str
            The file name for saving the network visualization image.
        """
        pos = nx.spring_layout(self.G)

        # Generate random colors for each community
        colors = [f'#{random.randint(0, 0xFFFFFF):06x}' for _ in range(len(self.unique_mofs))]

        plt.figure(figsize=(12, 10))

        for i, community_nodes in enumerate(self.unique_mofs):
            node_color = colors[i]
            nx.draw_networkx_nodes(self.G, pos, nodelist=community_nodes, node_size=50, node_color=node_color)

        # Draw edges and labels
        nx.draw_networkx_edges(self.G, pos, alpha=0.5)
        nx.draw_networkx_labels(self.G, pos, font_size=8)

        plt.title('MOF Network with Centrality and Communities')

        # Save the figure
        plt.savefig(output_image, dpi=300, bbox_inches='tight')
        plt.show()

    def save_unique_mofs_as_json(self, output_json='unique_mofs.json'):
        """
        Save the unique MOFs and their community groupings as a JSON file.

        Parameters:
        ----------
        output_json : str
            The file name for saving the unique MOFs as a JSON file.
        """
        mof_dict = {f'Community {i+1}': community for i, community in enumerate(self.unique_mofs)}

        # Save the dictionary as a JSON file
        with open(output_json, 'w') as json_file:
            json.dump(mof_dict, json_file, indent=4)

    def run_analysis(self):
        """
        Run the complete MOF network analysis: load data, create graph,
        compute centrality,
        detect communities, visualize the network, and save the results.
        """
        self.load_data()
        self.create_graph()
        centrality_metrics = self.compute_centrality()
        self.detect_communities()
        self.visualize_network(centrality_metrics)
        self.save_unique_mofs_as_json()



if __name__ == "__main__":
    analysis = MOFNetworkAnalysis('../../data/json_datai3.db', similarity_threshold=0.5)
    analysis.run_analysis()



#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################
import os
import pandas as pd
import argparse
import requests
import pubchempy as pcp
from mofstructure import mofdeconstructor
from fairmofsyncondition.read_write import coords_library



def pubchem_to_inchikey(filename):
    '''
    A function that retrieves chemical properties from PubChem using the given identifier.

    This function queries the PubChem database via the PubChemPy package and extracts
    a set of cheminformatic properties for the first matching compound. The properties
    include InChIKey, CID, IUPAC name, canonical SMILES, hydrogen bond donor and acceptor counts,
    rotatable bond count, and charge.

    Parameters:
        identifier (str): The chemical identifier to query (e.g., chemical name or SMILES).
        search_type (str): The type of identifier provided. Options include 'name' or 'smiles'.
                           Default is 'name'.

    Returns:
        dict or None: A dictionary containing the chemical properties if a matching compound is found;
                      otherwise, None.
    '''
    properties = {}
    ase_atom = coords_library.load_data_as_ase(filename)
    smile, inchi, inchikey = file2smile(filename)
    print(smile)
    print(inchikey)
    pubchem = pcp.get_compounds(smile, 'smiles')
    properties['inchikey'] = inchikey
    properties['smiles'] = smile
    properties['inchi'] = inchi

    if len(pubchem) > 0:
        all_prop = pubchem[0]
        properties['cid'] = all_prop.cid if all_prop.cid else None
        properties['iupac_name'] = all_prop.iupac_name if all_prop.iupac_name else None
        properties['h_bond_donor_count'] = all_prop.h_bond_donor_count if all_prop.h_bond_donor_count else None
        properties['h_bond_acceptor_count'] = all_prop.h_bond_acceptor_count if all_prop.h_bond_acceptor_count else None
        properties['rotatable_bond_count'] = all_prop.rotatable_bond_count if all_prop.rotatable_bond_count else None
        properties['charge'] = all_prop.charge if all_prop.charge else None
        # properties['synonyms'] = all_prop.synonyms if all_prop.synonyms else None
        # properties['sids'] = all_prop.sids if all_prop.sids else None
        return properties
    else:
        return None


def file2smile(filename):
    """
    Function that reads a filename and uses openbabel to compute smi, inChi and inChiKey
    **Parameters:**
        filename (str): name of file containing the structure. It can be in any ase readable
        file format as well as qchem out, AMS out, Gaussian out.

    **Return:**
        smi (str): SMILE strings
        inchi : inchi hashing of structure
        inChiKey : 28 character hashing


    """
    ase_atom = coords_library.load_data_as_ase(filename)
    smi, inChi, inChiKey = mofdeconstructor.compute_openbabel_cheminformatic(ase_atom)
    return smi, inChi, inChiKey

def print_helpful_information():
    """
    Prints helpful information about using the chemical_parser script.
    """
    help_text = """
        Usage: chemical_parser.py [CHEMICAL_IDENTIFIER] [OPTIONS]

        This script takes a chemical identifier as input (either a chemical name or a SMILES string),
        retrieves cheminformatic properties from PubChem (e.g., InChIKey, CID, IUPAC name, SMILES, etc.),
        and writes the information to a CSV file.

        Positional Arguments:
        File Name     Name of file containing structures.
                    ['any ase readable file', 'qchem.out', 'ams.out'] (Only non periodic systems)).

        Optional Arguments:
        -o, --output FILE        The path to the output CSV file (default: cheminfor.csv).


        Examples:
        struct2iupac water.xyz  -o water_info.csv
        struct2iupac  benzene.out
    """
    print(help_text)

def main():
    """
    Main function to parse command-line arguments, retrieve chemical information from PubChem,
    and write (or append) the data to a CSV file.
    """
    parser = argparse.ArgumentParser(
        description="Parse a chemical identifier and write cheminformatic properties to a CSV file."
    )
    parser.add_argument(
        'filename',
        type=str,
        nargs='?',
        default=None,
        help='filename of the molecule to determine name'
    )

    parser.add_argument(
        '-o', '--output',
        type=str,
        default='cheminfor.csv',
        help='The path to the output CSV file (default: cheminfor.csv).'
    )

    args = parser.parse_args()

    filename =  args.filename

    output_csv = args.output

    if filename is None:
        print_helpful_information()
        return

    chemical_info = pubchem_to_inchikey(filename)

    if chemical_info:
        df = pd.DataFrame([chemical_info])
        # If the output file exists, append without the header; otherwise, create a new file with the header.
        if os.path.exists(output_csv):
            df.to_csv(output_csv, mode='a', index=False, header=False)
            print(f"Chemical information appended to {output_csv}")
        else:
            df.to_csv(output_csv, index=False)
            print(f"Chemical information written to {output_csv}")
    else:
        print("No chemical information found.")

if __name__ == '__main__':
    main()


#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"
import os
import re
import pickle
import csv
import json
import codecs
from zipfile import ZipFile
import ase
import numpy as np
import pandas as pd
from ase import Atoms
import msgpack


class AtomsEncoder(json.JSONEncoder):
    '''
    Custom JSON encoder for serializing ASE `Atoms` objects and related data.

    This encoder converts ASE `Atoms` objects into JSON-serializable dictionaries.
    It also handles the serialization of ASE `Spacegroup` objects.

    **Methods**
    default(obj)
        Serializes objects that are instances of ASE `Atoms` or `Spacegroup`,
        or falls back to the default JSON encoder for unsupported types.

    **Examples**
        >>> from ase import Atoms
        >>> import json
        >>> atoms = Atoms('H2O', positions=[[0, 0, 0], [0, 0.76, 0], [0.76, 0, 0]])
        >>> json_data = json.dumps(atoms, cls=AtomsEncoder)
        >>> print(json_data)
    '''

    def default(self, encorder_obj):
        '''
        define different encoder to serialise ase atom objects
        '''
        if isinstance(encorder_obj, Atoms):
            coded = dict(positions=[list(pos) for pos in encorder_obj.get_positions()], lattice_vectors=[
                         list(c) for c in encorder_obj.get_cell()], labels=list(encorder_obj.get_chemical_symbols()))
            if len(encorder_obj.get_cell()) == 3:
                coded['periodic'] = ['True', 'True', 'True']
            coded['n_atoms'] = len(list(encorder_obj.get_chemical_symbols()))
            coded['atomic_numbers'] = encorder_obj.get_atomic_numbers().tolist()
            keys = list(encorder_obj.info.keys())
            if 'atom_indices_mapping' in keys:
                info = encorder_obj.info
                coded.update(info)
            return coded
        if isinstance(encorder_obj, ase.spacegroup.Spacegroup):
            return encorder_obj.todict()
        return json.JSONEncoder.default(self, encorder_obj)


def json_to_aseatom(data, filename):
    '''
    Serialize an ASE `Atoms` object and write it to a JSON file.
    This function uses the custom `AtomsEncoder` to convert an ASE `Atoms` object
    into a JSON format and writes the serialized data to the specified file.

    **parameters**
        data : Atoms or dict
            The ASE `Atoms` object or dictionary to serialize.
        filename : str
            The path to the JSON file where the serialized data will be saved.
    '''
    encoder = AtomsEncoder
    with open(filename, 'w', encoding='utf-8') as f_obj:
        json.dump(data, f_obj, indent=4, sort_keys=False, cls=encoder)
    return


def get_section(contents, start_key, stop_key, start_offset=0, stop_offset=0):
    """
    Extracts a section of lines from a list of strings between specified start and stop keys.
    This function searches through a list of strings (e.g., file contents) to find the last occurrence
    of a start key and extracts all lines up to and including the first occurrence of a stop key,
    with optional offsets for flexibility.

    **parameters**
        contents : list of str
            A list of strings representing the lines of a file or text content.
        start_key : str
            The key string that marks the start of the section.
        stop_key : str
            The key string that marks the end of the section.
        start_offset : int, optional
            The number of lines to include before the start key. Default is 0.
        stop_offset : int, optional
            The number of lines to include after the stop key. Default is 0.

    **returns**
        list of str
            The extracted lines from `contents` between the start and stop keys, including the offsets.
    """
    all_start_indices = []
    for i, line in enumerate(contents):
        if start_key in line:
            all_start_indices.append(i + start_offset)
    start_index = all_start_indices[-1]
    for i in range(start_index, len(contents)):
        line = contents[i]
        if stop_key in line:
            stop_index = i + 1 + stop_offset
            break
    data = contents[start_index:stop_index]
    return data


def append_json_atom(data, filename):
    '''
    Appends or updates a JSON file with data containing an ASE `Atoms` object.
    If the file does not exist or is empty, it creates a new JSON file with an empty dictionary
    as the initial content. The function then updates the file with the provided data using the
    custom `AtomsEncoder` for serializing ASE `Atoms` objects.

    **parameters**
        data : dict
            A dictionary containing data with an ASE `Atoms` object or other serializable content.
        filename : str
            The path to the JSON file where the data will be appended.
    '''
    encoder = AtomsEncoder
    if not os.path.exists(filename):
        with open(filename, 'w', encoding='utf-8') as f_obj:
            f_obj.write('{}')
    elif os.path.getsize(filename) == 0:
        with open(filename, 'w', encoding='utf-8') as f_obj:
            f_obj.write('{}')
    with open(filename, 'r+', encoding='utf-8') as f_obj:
        file_data = json.load(f_obj)
        file_data.update(data)
        f_obj.seek(0)

        json.dump(data, f_obj, indent=4, sort_keys=True, cls=encoder)


def numpy_to_json(ndarray, file_name):
    '''
    Serializes a NumPy array and saves it to a JSON file.
    This function converts a NumPy array into a list format, which is JSON-serializable,
    and writes it to the specified file.

    **parameters**
        ndarray : numpy.ndarray
            The NumPy array to serialize.
        file_name : str
            The path to the JSON file where the serialized data will be saved.
    '''
    json.dump(ndarray.tolist(), codecs.open(file_name, 'w',
              encoding='utf-8'), separators=(',', ':'), sort_keys=True)
    return


def list_2_json(list_obj, file_name):
    '''
    Writes a list to a JSON file.
    This function serializes a Python list and saves it to a
    specified JSON file.

    **parameters**
        list_obj : list
            The list to serialize and write to the file.
        file_name : str
            The path to the JSON file where the list will be saved.
    '''
    json.dump(list_obj, codecs.open(file_name, 'w', encoding='utf-8'))


def write_json(json_obj, file_name):
    '''
    Writes a Python dictionary object to a JSON file.

    This function serializes a Python dictionary into JSON
    format and writes it to the specified file and ensures that the
    JSON is human-readable with proper indentation.

    **parameters**
        json_obj : dict
            The Python dictionary to serialize and write to the JSON file.
        file_name : str
            The path to the JSON file where the data will be saved.
    '''
    json_object = json.dumps(json_obj, indent=4, sort_keys=True)
    with open(file_name, "w", encoding='utf-8') as outfile:
        outfile.write(json_object)


def json_to_numpy(json_file):
    '''
    Deserializes a JSON file containing a NumPy array
    back into a NumPy array. This function reads a JSON file,
    deserializes the data, and converts it into a NumPy array.

    **parameters**
        json_file : str
            The path to the JSON file containing the serialized NumPy array.

    **returns**
        numpy.ndarray
            The deserialized NumPy array.
    '''
    json_reader = codecs.open(json_file, 'r', encoding='utf-8').read()
    json_reader = np.array(json.loads(json_reader))
    return read_json


def append_json(new_data, filename):
    '''
    Appends new data to an existing JSON file. If the file does
    not exist or is empty, it creates a new JSON file with an
    empty dictionary. The function then updates the file with the
    provided data, overwriting existing keys if they are already present.

    **parameters**
        new_data : dict
            A dictionary containing the new data to append to the JSON file.
        filename : str
            The path to the JSON file.
    '''
    if not os.path.exists(filename):
        with open(filename, 'w', encoding='utf-8') as file:
            file.write('{}')
    elif os.path.getsize(filename) == 0:
        with open(filename, 'w', encoding='utf-8') as file:
            file.write('{}')
    with open(filename, 'r+', encoding='utf-8') as file:
        file_data = json.load(file)
        file_data.update(new_data)
        file.seek(0)
        json.dump(file_data, file, indent=4, sort_keys=True)


def read_json(file_name):
    '''
    Loads and reads a JSON file. This function
    opens a JSON file, reads its content, and deserializes it into
    a Python object (e.g., a dictionary or list).

    **Parameters**
        file_name : str
            The path to the JSON file to be read.

    **returns**
        dict or list
            The deserialized content of the JSON file.
    '''
    with open(file_name, 'r', encoding='utf-8') as f_obj:
        data = json.load(f_obj)

    return data


def csv_read(csv_file):
    '''
    Reads a CSV file and returns its content as a list of rows. This function reads
    the content of a CSV file and returns it as a list.

    **parameters**
        csv_file : str
            The path to the CSV file to be read.

    **returns**
        list of list of str
            A list of rows from the CSV file. Each row is a list of strings.

    '''
    f_obj = open(csv_file, 'r', encoding='utf-8')
    data = csv.reader(f_obj)
    return data


def get_contents(filename):
    '''
    Reads the content of a file and returns it as a list of lines.
    This function opens a file, reads its content line by line,
    and returns a list where each element is a line from the file,
    including newline characters.

    **parameters**
        filename : str
            The path to the file to be read.

    **returns**
        list of str
            A list containing all lines in the file.
    '''
    with open(filename, 'r', encoding='utf-8') as f_obj:
        contents = f_obj.readlines()
    return contents


def put_contents(filename, output):
    '''
    Writes a list of strings into a file. This function writes the content of a list to a file, where each element
    in the list represents a line to be written. If the file already exists,
    it will be overwritten.

    **parameters**
        filename : str
            The path to the file where the content will be written.
        output : list of str
            A list of strings to be written to the file. Each string represents
            a line, and newline characters should be included if needed.
    '''
    with open(filename, 'w', encoding='utf-8') as f_obj:
        f_obj.writelines(output)
    return


def append_contents(filename, output):
    '''
    Appends a list of strings to a file. This function appends
    the content of a list to a file, where each element in the
    list represents a line to be written. If the file does not exist,
    it will be created.

    **parameters**
        filename : str
            The path to the file where the content will be appended.
        output : list of str
            A list of strings to be appended to the file. Each string represents
            a line, and newline characters should be included if needed.
    '''
    with open(filename, 'a', encoding='utf-8') as f_obj:
        f_obj.writelines(output)
    return


def save_pickle(model, file_path):
    '''
    Saves a Python object to a file using pickle. This function serializes
    a Python object and saves it to a specified file
    in binary format using the `pickle` module.

    **parameters**
        model : object
            The Python object to serialize and save.
        file_path : str
            The path to the file where the object will be saved.
    '''
    with open(file_path, 'wb') as file:
        pickle.dump(model, file)


def append_pickle(new_data, filename):
    '''
    Appends new data to a pickle file. This function appends new data to an existing pickle file. If the file does not
    exist, it will be created. Data is appended in binary format, ensuring that
    previously stored data is not overwritten.

    **parameters**
        new_data : object
            The Python object to append to the pickle file.
        filename : str
            The path to the pickle file where the data will be appended.
    '''
    with open(filename, 'ab') as f_:
        pickle.dump(new_data, f_)
    f_.close()


def pickle_load(filename):
    '''
    Loads and deserializes data from a pickle file. This
    function reads a pickle file and deserializes its content into
    a Python object.

    **parameters**
        filename : str
            The path to the pickle file to be loaded.

    **returns**
        object
            The deserialized Python object from the pickle file
    '''
    data = open(filename, 'rb')
    data = pickle.load(data)
    return data


def read_zip(zip_file):
    '''
    Reads and extracts the contents of a zip file.

    This function opens a zip file and extracts its
    contents to the specified directory. If no directory
    is provided, it extracts to the current working
    directory.

    **parameters**
        zip_file : str
            The path to the zip file to be read and extracted.
        extract_to : str, optional
            The directory where the contents of the zip file will be extracted.
            If not provided, the current working directory is used.

    **returns**
        list of str
            A list of file names contained in the zip file.
    '''
    content = ZipFile(zip_file, 'r')
    content.extractall(zip_file)
    content.close()
    return content


def remove_trailing_commas(json_file):
    '''
    Cleans trailing commas in a JSON file and returns
    the cleaned JSON string. This function reads a JSON file,
    removes trailing commas from objects and arrays,
    and returns the cleaned JSON string. It is useful
    for handling improperly formatted JSON files with
    trailing commas that are not compliant with the JSON standard.

    **parameters**
        json_file : str
            The path to the JSON file to be cleaned.

    **returns**
        cleaned_json str
            A cleaned JSON string with trailing commas removed.

    '''
    with open(json_file, 'r', encoding='utf-8') as file:
        json_string = file.read()

    trailing_object_commas_re = re.compile(r',(?!\s*?[\{\[\"\'\w])\s*}')
    trailing_array_commas_re = re.compile(r',(?!\s*?[\{\[\"\'\w])\s*\]')

    objects_fixed = trailing_object_commas_re.sub("}", json_string)
    cleaned_json = trailing_array_commas_re.sub("]", objects_fixed)

    return cleaned_json


def query_data(ref, data_object, col=None):
    '''
    Queries data from a CSV (as a DataFrame) or JSON (as a dictionary).

    This function retrieves data based on a reference key or value from either
    a dictionary (JSON-like object) or a pandas DataFrame (CSV-like object).

    **parameters**
        ref : str or int
            The reference key or value to query.
        data_object : dict or pandas.DataFrame
            The data source, which can be a dictionary (for JSON) or a pandas
            DataFrame (for CSV).
        col : str, optional
            The column name to query in the DataFrame. This parameter is
            required if the data source is a DataFrame and ignored if
            the data source is a dictionary.

    **returns**
        object
            The queried data. For a dictionary, it returns the value
            associated with the reference key.
            For a DataFrame, it returns the rows
            where the specified column matches the reference value.
    '''
    if isinstance(data_object, dict):
        return data_object[ref]
    else:
        return data_object.loc[data_object[col] == ref]


def combine_json_files(file1_path, file2_path, output_path):
    '''
    Queries data from a CSV (as a DataFrame) or JSON (as a dictionary).

    This function retrieves data based on a reference key or value from either
    a dictionary (JSON-like object) or a pandas DataFrame (CSV-like object).

    **parameters**
        ref : str or int
            The reference key or value to query.
        data_object : dict or pandas.DataFrame
            The data source, which can be a dictionary (for JSON) or a pandas
            DataFrame (for CSV).
        col : str, optional
            The column name to query in the DataFrame. This parameter is
            required if the data source
            is a DataFrame and ignored if the data source is a dictionary.

    **returns**
        object
            The queried data. For a dictionary, it returns the value
            associated with the reference key.
            For a DataFrame, it returns the rows where the specified
            column matches the reference value.
    '''
    with open(file1_path, 'r', encoding='utf-8') as file1:
        data1 = json.load(file1)

    with open(file2_path, 'r', encoding='utf-8') as file2:
        data2 = json.load(file2)

    combined_data = {**data1, **data2}

    with open(output_path, 'w', encoding='utf-8') as output_file:
        json.dump(combined_data, output_file, indent=2)


def save_dict_msgpack(data: dict, filename: str) -> None:
    """Save a dictionary to a file using MessagePack."""
    with open(filename, "wb") as f:
        msgpack.pack(data, f, use_bin_type=True)


def load_dict_msgpack(filename: str) -> dict:
    """Load a dictionary from a MessagePack file."""
    with open(filename, "rb") as f:
        return msgpack.unpack(f, raw=False, strict_map_key=False)


def load_data(filename):
    '''
    Automatically detects the file extension and loads the data using the
    appropriate function. This function reads a file and returns
    its content, choosing the correct loading method based on the file
    extension. Supported file formats include JSON, CSV, Pickle, Excel,
    and plain text files.

    **parameters**
        filename : str
            The path to the file to be loaded.

    **returns**
        object
            The loaded data, which can be a dictionary, DataFrame, list, or other Python object,
            depending on the file type.
    '''
    file_ext = filename[filename.rindex('.')+1:]
    if file_ext == 'json':
        data = read_json(filename)
    elif file_ext == 'csv':
        data = pd.read_csv(filename)
    elif file_ext == 'p' or file_ext == 'pkl':
        data = pickle_load(filename)
    elif file_ext == 'xlsx':
        data = pd.read_excel(filename)
    elif file_ext == 'msgpack':
        data = load_dict_msgpack(filename)
    else:
        data = get_contents(filename)
    return data



#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################
from ase.io import read
import warnings
import numpy as np
from scipy.signal import find_peaks
from lmfit.models import GaussianModel, PolynomialModel  #
from pymatgen.analysis.diffraction.xrd import XRDCalculator
from pymatgen.analysis.diffraction.neutron import NDCalculator
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core import Structure
from scipy.optimize import curve_fit
from scipy.special import erf, wofz
from scipy.stats import linregress
from scipy.optimize import OptimizeWarning
warnings.filterwarnings("ignore", category=OptimizeWarning)


class Crystallinity(object):
    def __init__(self,
                 ase_atoms=None,
                 filename=None,
                 wavelength='CuKa',
                 diffraction_type="PXRD"
                 ):
        """
        Compute the crystallinity of a crystal structure.

        Parameters:
            atoms (ase.Atoms): Crystal structure.

        Returns:
            float: Crystallinity.
        """
        if ase_atoms is not None:
            self.ase_atoms = ase_atoms
            self.structure = AseAtomsAdaptor.get_structure(self.ase_atoms)
        else:
            with open(filename, "rt", encoding="utf-8") as f:
                cif_data = f.read()
            self.structure = Structure.from_str(cif_data, fmt="cif")


        self.wavelength = wavelength
        self.diffraction_type = diffraction_type

    def find_diffraction_pattern(self):
        """
        Compute the diffraction pattern of the crystal structure.

        Returns:
            list: Diffraction pattern.
        """
        if self.diffraction_type == "PXRD":
            xrd = XRDCalculator(wavelength=self.wavelength)
            pattern = xrd.get_pattern(self.structure)
        elif self.diffraction_type == "ND":
            xrd = NDCalculator(wavelength=self.wavelength)
            pattern = xrd.get_pattern(self.structure)
        else:
            raise ValueError("""Invalid diffraction type.
                             Choose between PXRD and ND""")

        return pattern

    def get_pattern(self):
        """
        """
        pattern = self.find_diffraction_pattern()
        two_theta = np.array(pattern.x)
        intensity = np.array(pattern.y)
        return two_theta, intensity


    def compute_crystallinity_area_method(self,
                                          peak_prominence=1.0,
                                          height_threshold=0.01,
                                          polynomial_order=2):
        """
        Estimate the Crystallinity Index (CI) via the area method using
        peak deconvolution and background (amorphous) modeling.

        Steps:
        1. Obtain diffraction pattern (2θ vs intensity).
        2. Model the amorphous background with a polynomial.
        3. Identify crystalline peaks (using find_peaks).
        4. Build a composite model: polynomial background + one Gaussian per peak.
        5. Fit the composite model to the data.
        6. Integrate the total area under the fitted curve.
        7. Integrate the area under just the background.
        8. Crystallinity Index = (Total area - background area) / (Total area).

        Parameters:
        peak_prominence (float): Sensitivity for peak detection.
        height_threshold (float): Minimum height threshold as a fraction of max intensity.
        polynomial_order (int): Degree of polynomial for background modeling.

        Returns:
        crystallinity_index (float): Fraction in [0,1] representing the crystalline area.
        """
        # 1. Get the raw diffraction data.
        # Assume self.get_pattern() returns (two_theta, intensity)
        two_theta, intensity = self.get_pattern()

        # Ensure intensities are nonnegative.
        intensity = np.clip(intensity, 0, None)

        # 2. (Optional) Crop the data range if desired.
        # two_theta, intensity = self._crop_data(two_theta, intensity, lower=5, upper=80)

        # To help the polynomial fit, shift two_theta by subtracting its mean.
        two_theta_mean = np.mean(two_theta)
        two_theta_shifted = two_theta - two_theta_mean  # Now values are centered near zero.

        # 3. Identify peaks.
        # Use an optimal prominence if available, otherwise use peak_prominence.
        try:
            optimal_prominence, _ = compute_optimal_prominence(intensity)
        except Exception:
            optimal_prominence = peak_prominence

        peaks, properties = find_peaks(intensity, prominence=optimal_prominence,
                                    height=height_threshold * np.max(intensity))
        peak_centers = two_theta[peaks]

        # 4. Build composite model: background (polynomial) + sum of Gaussians.
        # For the background, use the shifted two_theta values.
        background_model = PolynomialModel(prefix='poly_', degree=polynomial_order, independent_vars=['x'])
        model = background_model
        pars = background_model.make_params(x=two_theta_shifted)

        # Set initial guess for the background constant as the minimum intensity.
        pars['poly_c0'].set(value=np.min(intensity), min=0, max=np.max(intensity))
        # For higher-order terms, set initial guesses to 0 with narrow bounds.
        for i in range(1, polynomial_order+1):
            key = f'poly_c{i}'
            pars[key].set(value=0, min=-np.max(intensity)*1e-3, max=np.max(intensity)*1e-3)

        # Add one Gaussian per detected peak.
        for i, cen in enumerate(peak_centers):
            gauss = GaussianModel(prefix=f'g{i}_', independent_vars=['x'])
            model = model + gauss

            init_pars = gauss.make_params(x=two_theta_shifted)
            # Adjust the center: shift by the same amount.
            shifted_cen = cen - two_theta_mean
            init_pars[f'g{i}_center'].set(value=shifted_cen, min=shifted_cen-0.5, max=shifted_cen+0.5)
            init_pars[f'g{i}_sigma'].set(value=0.2, min=0.01, max=2.0)
            # Use the intensity at the detected peak as the amplitude.
            peak_height = intensity[peaks[i]]
            init_pars[f'g{i}_amplitude'].set(value=peak_height, min=0)

            pars.update(init_pars)

        # 5. Fit the composite model.
        try:
            result = model.fit(intensity, pars, x=two_theta_shifted)
        except Exception as e:
            raise ValueError(f"Model fitting failed: {e}")

        # Check if the fitted model returns any NaN values.
        fitted_intensity = result.eval(x=two_theta_shifted)
        if np.any(np.isnan(fitted_intensity)):
            raise ValueError("""Fitted model produced NaN values.
                             Check initial guesses or parameter bounds.
                             """
                             )

        # 6. Compute total area under the fitted composite model (in original two_theta).
        # Here we evaluate the fitted function on the original two_theta (undo the shift).
        total_area = np.trapz(result.eval(x=two_theta - two_theta_mean), two_theta)

        # 7. Compute the background area using the polynomial model.
        poly_pars = result.params.copy()
        # Zero out all Gaussian parameters.
        for name in poly_pars:
            if name.startswith('g'):
                poly_pars[name].set(value=0)
        background_only = background_model.eval(params=poly_pars, x=two_theta_shifted)
        background_area = np.trapz(background_only, two_theta)

        # 8. Crystalline area is the total area minus the background area.
        crystalline_area = total_area - background_area
        crystallinity_index = crystalline_area / total_area if total_area != 0 else 0.0

        return crystallinity_index

    def _crop_data(self, x, y, lower=5, upper=80):
        """Helper to crop data between lower and upper 2theta values."""
        mask = (x >= lower) & (x <= upper)
        return x[mask], y[mask]

    def get_size_and_strain(self, wavelength=0.15406, k_value=0.9):
        """

        """
        sizes = []
        strains = []
        two_theta, intensity = self.get_pattern()
        fwhm_data, peak_positions = estimate_fwhm_from_pxrd(two_theta, intensity)
        for fwhm, center in zip(fwhm_data, peak_positions):
            theta = center / 2.0
            size, strain = compute_crystallite_size_and_strain(theta, fwhm, wavelength=wavelength, k_value=k_value)
            sizes.append(size)
            strains.append(strain)
        return sizes, strains

    def get_average_size_and_strain(self,
                                    wavelength=0.15406,
                                    k_value=0.9
                                    ):
        """
        """
        sizes, strains = self.get_size_and_strain(wavelength=wavelength, k_value=k_value)
        return np.mean(sizes), np.mean(strains)

    def get_modified_scherrer(self,
                              wavelength=0.15406,
                              k_value=0.9
                              ):
        two_theta, intensity = self.get_pattern()
        fwhms, peak_positions = estimate_fwhm_from_pxrd(two_theta, intensity)
        size = modified_scherrer_eq(fwhms, peak_positions, wavelength=wavelength, k_value=k_value)
        return size



def voigt_profile(x,
                  amplitude,
                  center,
                  sigma,
                  gamma):
    """
    Voigt profile function using the Faddeeva function.
    """
    z = ((x - center) + 1j * gamma) / (sigma * np.sqrt(2))
    return amplitude * np.real(wofz(z)) / (sigma * np.sqrt(2 * np.pi))


def compute_optimal_prominence(intensity,
                               min_prominence=0.01,
                               max_prominence=1.0,
                               tolerance=3,
                               step=0.01):
    """
    Automatically determines the optimal prominence for peak detection.
    """
    max_prominence_avg = 0
    optimal_prominence = min_prominence
    no_improvement_count = 0

    if np.max(intensity) == 0:
        raise ValueError("Maximum intensity is zero, unable to normalize.")

    norm_intensity = intensity / np.max(intensity)

    for current_prominence in np.arange(min_prominence, max_prominence, step):
        peaks, properties = find_peaks(norm_intensity, prominence=current_prominence)
        prominences = properties["prominences"]
        if len(prominences) > 0:
            current_avg_prominence = np.mean(prominences)
            if current_avg_prominence > max_prominence_avg:
                max_prominence_avg = current_avg_prominence
                optimal_prominence = current_prominence
                no_improvement_count = 0
            else:
                no_improvement_count += 1
        if no_improvement_count >= tolerance:
            break
    return optimal_prominence, peaks


def estimate_fwhm_from_pxrd(two_theta,
                            intensities,
                            height_threshold=0.01
                            ):
    """
    Estimate the FWHM values and peak positions from PXRD data.

    This function detects peaks, then for each peak it extracts the region and fits a Voigt
    profile to determine an accurate FWHM.

    Parameters:
        two_theta : np.array
            Array of 2θ values (in degrees).
        intensities : np.array
            Array of intensity values corresponding to the 2θ values.
        height_threshold : float, optional
            Minimum height threshold as a fraction of the maximum intensity.

    Returns:
        fwhm_data : np.array
            Estimated FWHM (in degrees) for each detected peak.
        peak_positions : np.array
            2θ positions (in degrees) of each detected peak.
    """
    # Determine an optimal prominence for the data
    optimal_prominence, _ = compute_optimal_prominence(intensities)
    peaks, properties = find_peaks(intensities, prominence=optimal_prominence,
                                   height=height_threshold * np.max(intensities))
    fwhm_data = []
    peak_positions = []

    for i, peak in enumerate(peaks):
        left_base = properties["left_bases"][i]
        right_base = properties["right_bases"][i]
        peak_region_x = two_theta[left_base:right_base]
        peak_region_y = intensities[left_base:right_base]

        amplitude = peak_region_y.max()
        center = two_theta[peak]
        half_max = amplitude / 2
        # Find indices where the intensity is close to half maximum
        closest_to_half_max = np.where(np.isclose(peak_region_y, half_max, atol=0.1 * half_max))[0]

        if len(closest_to_half_max) >= 2:
            estimated_fwhm = peak_region_x[closest_to_half_max[-1]] - peak_region_x[closest_to_half_max[0]]
            sigma = estimated_fwhm / (2 * np.sqrt(2 * np.log(2)))
        else:
            estimated_fwhm = (peak_region_x[-1] - peak_region_x[0]) / 2
            sigma = estimated_fwhm / (2 * np.sqrt(2 * np.log(2)))

        gamma = estimated_fwhm / 2 if 'estimated_fwhm' in locals() else 0.1

        try:
            popt, _ = curve_fit(voigt_profile, peak_region_x, peak_region_y,
                                p0=[amplitude, center, sigma, gamma],
                                method='trf', max_nfev=5000,
                                bounds=([0, center - 2, 0, 0], [np.inf, center + 2, np.inf, np.inf]))
            fitted_sigma = popt[2]
            fitted_gamma = popt[3]
            # Compute FWHM for Voigt profile using an approximate relation
            fwhm = 0.5346 * (2 * fitted_gamma) + np.sqrt(0.2166 * (2 * fitted_gamma)**2 + (2 * fitted_sigma)**2)
        except RuntimeError:
            fwhm = estimated_fwhm

        fwhm_data.append(fwhm)
        peak_positions.append(center)

    return np.array(fwhm_data), np.array(peak_positions)

def compute_crystallite_size_and_strain(theta_deg, fwhm_deg, wavelength=0.15406, k_value=0.9):
    """
    Compute the crystallite size (D) and microstrain (ε) for a given peak.

    Parameters:
        theta_deg : float
            Bragg angle (θ in degrees, half of the 2θ value).
        fwhm_deg : float
            FWHM (in degrees, as estimated from the PXRD pattern).
        wavelength : float
            X-ray wavelength in nm.
        k : float
            Scherrer constant.

    Returns:
        D : float
            Crystallite size in nm.
        strain : float
            Microstrain (dimensionless).
    """
    theta = np.deg2rad(theta_deg)
    beta = np.deg2rad(fwhm_deg)
    size = (k_value * wavelength) / (beta * np.cos(theta))
    strain = beta / (4 * np.tan(theta))
    return size, strain


def modified_scherrer_eq(fwhms,
                         two_theta_position,
                         wavelength=0.15406,
                         k_value=0.9,
                         fwhm_in_degrees=True
                         ):
    """
    Computes the crystallite size using the modified Scherrer equation.

    The modified Scherrer equation is based on:

        Kλ / L = β cosθ

    Taking logarithms gives:

        ln(β) = ln(1/cosθ) + ln(Kλ/L)

    A linear regression of ln(β) versus ln(1/cosθ) provides an intercept = ln(Kλ/L),
    from which the crystallite size L is calculated as:

        L = (K * wavelength) / exp(intercept)

    Parameters:
    ----------
    fwhms : array-like
        Array of FWHM values for the peaks. If these are in degrees,
        set fwhm_in_degrees=True (default).
    two_theta_position : array-like
        Array of 2θ values (in degrees) for the peaks.
    wavelength : float, optional
        X-ray wavelength in nm (default: 0.15406 nm).
    k_value : float, optional
        Shape factor (default: 0.9).
    fwhm_in_degrees : bool, optional
        If True, converts fwhms from degrees to radians (default: True).

    Returns:
    -------
    crystallite_size_modified : float
        Calculated crystallite size (L) in nm.
    """
    # Convert inputs to numpy arrays
    fwhms = np.array(fwhms)
    two_theta_position = np.array(two_theta_position)

    # Convert FWHM to radians if provided in degrees
    if fwhm_in_degrees:
        fwhms = np.radians(fwhms)

    # Calculate ln(β) where β is in radians
    ln_beta = np.log(fwhms)

    # Compute θ (half of 2θ) in radians
    theta = np.radians(two_theta_position) / 2.0
    # Calculate ln(1/cosθ)
    ln_1_cos_theta = np.log(1 / np.cos(theta))

    # Perform linear regression: ln(β) = m * ln(1/cosθ) + intercept
    slope, intercept, r_value, p_value, std_err = linregress(ln_1_cos_theta, ln_beta)

    # The intercept is ln(K * wavelength / L), so:
    # L = (K * wavelength) / exp(intercept)
    crystallite_size_modified = (k_value * wavelength) / np.exp(intercept)

    return crystallite_size_modified




#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################


import torch
import numpy as np
from sentence_transformers import SentenceTransformer, util


def onehot_encoder(data, category):
    '''
    A function that performs one-hot encoding on a given categorical feature.

    This function takes a categorical feature and
    performs one-hot encoding to convert
    it into a numerical representation.
    The function returns a binary matrix where each
    column corresponds to a unique category in the feature.

    Parameters:
        data (list): The categorical feature to encode.
        category (list): The unique categories in the feature.

    Returns:
        numpy.ndarray: A binary matrix representing the
        one-hot encoded feature.
    '''
    onehot = np.zeros((len(data), len(category)))
    for i, cat in enumerate(data):
        onehot[i][category.index(cat)] = 1
    return onehot


def onehot_encoder_pyg(data, category):
    '''
    A function that performs one-hot encoding on a given categorical feature
    and returns a PyTorch tensor compatible with PyTorch Geometric.

    Parameters:
        data (list): The categorical feature to encode.
        category (list): The unique categories in the feature.

    Returns:
        torch.Tensor: A binary tensor representing the
        one-hot encoded feature.
    '''
    onehot = torch.zeros((len(data), len(category)), dtype=torch.float32)
    for i, cat in enumerate(data):
        onehot[i, category.index(cat)] = 1
    return onehot

def reverse_onehot_pyg(onehot_tensor, category):
    '''
    Converts a one-hot encoded PyTorch tensor back to categorical labels.

    Parameters:
        onehot_tensor (torch.Tensor): The one-hot encoded tensor.
        category (list): The original categories corresponding to indices.

    Returns:
        list: The original categorical values.
    '''
    indices = torch.argmax(onehot_tensor, dim=1)
    return [category[i] for i in indices]


def load_sentence_model(list_of_words):
    """
    Function to load a word embedding to facilitate data retrieval

    Parameters:
        list_of_words (list): A list of textual data.

    Returns:
        dict_embeddings (torch.Tensor): A tensor of
        embeddings for the dictionary.

    """
    model = SentenceTransformer('all-MiniLM-L6-v2')
    dict_embeddings = model.encode(list_of_words, convert_to_tensor=True)
    return dict_embeddings, model


def best_embedding_match(query, dictionary_keys, dict_embeddings, model):
    """
    Function to find the most similar embedding in a dictionary
    based on cosine similarity between the query embedding and the embeddings
    in the dictionary.

    Parameters:
        query (str): The text input query.
        dictionary_keys (list): A list of keys in the dictionary.
        dict_embeddings (torch.Tensor): A tensor of embeddings
        for the dictionary.
        model (SentenceTransformer): The SentenceTransformer model
        for embedding generation.

    Returns:
        str: The key from the dictionary that has the highest
        cosine similarity with the query embedding, or the original
        query if no match is found.

    """
    query_emb = model.encode(query, convert_to_tensor=True)
    scores = util.cos_sim(query_emb, dict_embeddings)[0]
    best_idx = torch.argmax(scores).item()
    best_score = scores[best_idx].item()
    print(best_score)
    if best_score >= 0.9:
        return dictionary_keys[best_idx]
    else:
        return query



import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
from matplotlib.colors import ListedColormap
from sklearn.neighbors import NearestNeighbors
from fairmofsyncondition.read_write import filetyper

def predict_kmeans(data, num_clusters, random_state=0):
    model = KMeans(n_clusters=num_clusters, random_state=random_state)
    model.fit(data)
    labels = model.predict(data)
    centroids = model.cluster_centers_
    return labels, centroids, model.inertia_


class Cluster:
    def __init__(self, data, random_state=0):
        self.data = data

    def find_number_of_clusters(self, path_to_save='number_of_clusters.png', show_plot=False):
        num_clusters = range(1, 30)
        inertias = []
        for k in num_clusters:
            _, _, inertia_ = self.predict_kmeans(self.data, k)
            inertias.append(inertia_)

        plt.figure(figsize=(8, 6))
        plt.plot(num_clusters, inertias, '-o')
        plt.xlabel('Number of Clusters (k)', fontsize=17)
        plt.ylabel('Inertia', fontsize=17)
        plt.tick_params(axis='both', which='major', labelsize=14)
        plt.savefig(path_to_save, dpi=600)
        if show_plot:
            plt.show()
        plt.clf()

    def plot_clustering(self, num_clusters, path_to_save='cluster_data', colors=None, marker='D'):
        if not os.path.exists(path_to_save):
            os.makedirs(path_to_save)

        scaler = StandardScaler()
        self.scaled_data = scaler.fit_transform(self.data)
        self.labels, self.centroids, _ = predict_kmeans(self.scaled_data, num_clusters)
        self.centroids_original = scaler.inverse_transform(self.centroids)
        if colors is None:
            colors = plt.cm.viridis(np.linspace(0, 1, num_clusters))

        plt.figure(figsize=(10, 6))
        plt.scatter(self.data[:, 0], self.data[:, 0], c=self.labels, cmap=ListedColormap(
            colors), alpha=0.7, marker='o')

        plt.scatter(self.centroids_original[:, 0], self.centroids_original[:, 0],
                    c='black', s=60, marker=marker, label='Centroids')

        plt.savefig(f'{path_to_save}/clusters_clusters.png', dpi=600)
        plt.show()
        plt.clf()

    def extract_neighbors(self, dict_value):
        json_data = {}
        neighbors_data = {}
        tmp = []
        tmp2 = {}
        nbrs = NearestNeighbors(n_neighbors=11).fit(self.data)
        for i, centroid in enumerate(self.centroids_original):
            distances = np.linalg.norm(self.data - centroid, axis=1)
            closest_index = np.argmin(distances)
            refcode_1 = dict_value[closest_index]
            tmp.append(
                {"refcode": refcode_1, 'energy': self.data[closest_index].tolist()[0]})
            neighbor_distances, indices = nbrs.kneighbors([centroid])
            neighbors = indices[0]
            tmp_neighbors = []
            for idx in neighbors:
                refcode = dict_value[idx]
                tmp_neighbors.append(
                    {"refcode": refcode, 'energy': self.data[idx].tolist()[0]})
            tmp2[i+1] = tmp_neighbors
        json_data.update(tmp)
        neighbors_data.update(tmp2)

energy_data = filetyper.load_data('../../data/compiled_ligand_bde_energy.json')

data = np.array(list(energy_data.values())).reshape(-1,1)
clustering = Cluster(data)
# clustering.find_number_of_clusters(show_plot=True)
clustering.plot_clustering(num_clusters=5)


#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################

import argparse
import random
import optuna
from torch import nn
import torch
from tqdm import tqdm
from torch_geometric.loader import DataLoader
from fairmofsyncondition.read_write import coords_library, filetyper
from fairmofsyncondition.model.thermodynamic_stability import  EnergyGNN_GAT2Conv, EnergyGNN_GIN, GraphLatticeModel, train, evaluate, load_dataset


def fine_opt_paramter(path_to_lmbd, mol_def):
    """
    Fine-tuning the optimization parameters using Optuna.

    Parameters:
        path_to_lmbd (str): Path to the lambda file.
        mol_def (str): Molecule definition ('GAT' or 'GIN').

    Returns:
        float: Validation loss.
    """
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    def objective(trial):
        """
        Objective function for hyperparameter optimization using Optuna.

        Parameters:
            trial (optuna.Trial): A trial object for sampling hyperparameters.

        Returns:
            float: Validation loss.
        """

        hidden_dim = trial.suggest_categorical("hidden_dim", [64, 128, 256])
        learning_rate = trial.suggest_categorical(
            "learning_rate", [1e-4, 1e-3, 1e-2, 1e-1])
        batch_size = trial.suggest_categorical("batch_size", [2048])
        dropout = trial.suggest_float("dropout", 0.1, 0.9)
        heads = trial.suggest_int("heads", 2, 6)
        # epoch = trial.suggest_int("epoch", 100,  5000)

        train_loader, val_loader, test_dataset, normalise_parameter = load_dataset(path_to_lmbd, batch_size=batch_size)
        if mol_def == 'GAT':
            model = EnergyGNN_GAT2Conv(input_dim=4, hidden_dim=hidden_dim, output_dim=1,
                          edge_dim=1, heads=heads, dropout=dropout).to(device)
        elif mol_def == 'GIN':
            model = EnergyGNN_GIN(input_dim=4, hidden_dim=hidden_dim, output_dim=1,
                          edge_dim=1, heads=heads, dropout=dropout).to(device)
        elif mol_def == 'lattice':
            lattice_hidden_dim = trial.suggest_categorical("lattice_hidden_dim",[ 5, 10, 15, 20, 25])
            num_layers= trial.suggest_categorical("num_layers",[ 2, 3, 4])
            model = GraphLatticeModel(input_dim=4,
                                gnn_hidden_dim=hidden_dim,
                                lattice_hidden_dim=lattice_hidden_dim,
                                output_dim=1,
                                num_layers=num_layers,
                                dropout=dropout).to(device)
        optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate, weight_decay=0.00001)
        # criterion = nn.MSELoss()
        criterion = nn.L1Loss()

        for epoch in tqdm(range(100)):
            train_loss = train(model, train_loader,
                               optimizer, criterion, device)

        val_loss = evaluate(model, val_loader, criterion, device)

        print(
            f"Trial: Hidden_dim={hidden_dim}, LR={learning_rate:.5f}, Batch_size={batch_size}, Heads={heads}, Dropout={dropout}, train_loss={train_loss:.4f} Val_loss={val_loss:.4f}")
        return val_loss

    study = optuna.create_study(direction="minimize")
    study.optimize(objective, n_trials=100)

    best_params = study.best_params
    print(f"Optimal parameters: {best_params}")
    print(f"Best validation loss: {study.best_value}")

    best_model_config = {
        "hidden_dim": best_params["hidden_dim"],
        "learning_rate": best_params["learning_rate"],
        "batch_size": best_params["batch_size"],
        "heads": best_params["heads"],
        "dropout": best_params["dropout"],
        "epoch": best_params["epoch"],
        "val_loss": study.best_value
    }
    if mol_def == 'lattice':
        best_model_config = {
        "hidden_dim": best_params["hidden_dim"],
        "learning_rate": best_params["learning_rate"],
        "batch_size": best_params["batch_size"],
        "heads": best_params["heads"],
        "dropout": best_params["dropout"],
        "epoch": best_params["epoch"],
        "lattice_hidden_dim": best_params["lattice_hidden_dim"],
        "num_layers": best_params["num_layers"],
        "val_loss": study.best_value
    }

    output_file = f"{mol_def}_best_model_config.txt"
    with open(output_file, "w") as f:
        for key, value in best_model_config.items():
            f.write(f"{key}: {value}\n")

    print(f"Best model configuration written to {output_file}")


def main():
    parser = argparse.ArgumentParser(
        description="Run hyperparameter optimization for EnergyGNN model.")
    parser.add_argument("--path_to_lmbd", type=str,
                        required=True, help="Path to the dataset directory")
    parser.add_argument("--mol_def", type=str,
                        required=True, choices=["GAT", "GIN", "lattice"],
                        help="Model definition: GAT, GIN, or lattice")
    args = parser.parse_args()
    fine_opt_paramter(args.path_to_lmbd, args.mol_def)


if __name__ == "__main__":
    main()



#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################

import torch


class SimilarityFinder:
    """
    A class to compute the similarity between two ASE Atoms graph systems.
    These  graphs corresponds to AtomGraph from orbital-materials module
    which provides and efficient way to convert ase atoms to graphs. This is
    particularly useful for converting ase crystal structures to graphs for
    directly comparison.

    This class takes two atom graph representations (graph1 and graph2),
    and computes a similarity index between them based on node features,
    edge features, and other properties of the graph.

    The approach used to compute similarity involves comparing key features of
    the two AtomGraphs. These features include the number of nodes (representing
    atoms), edges (representing bonds or connections between atoms), atomic numbers,
    node embeddings, and edge properties such as vectors and the relationships between
    senders and receivers of these edges. The senders and receivers here correspond to
    the indices of atoms in the graph.

    - The sender is the starting atom of a bond
    - The receiver is the atom that the bond points to

    Moreover, incases where features of both graph differe in size, a padding strategy
    is applied to equalize their lengths to optimise the comparison between the graphs.
    Each feature difference is normalized, weighted based
    on its importance, and then combined to form a final similarity index that ranges
    from 0 to 1.
    - 1 indicates perfect similarity.
    - 0 indicates no similarity.
    """

    def __init__(self, graph1, graph2):
        """
        Initialize the SimilarityFinder with two graphs.

        **parameters:**
            - graph1: The first atom graph (an instance of an AtomGraph or similar structure).
            - graph2: The second atom graph (an instance of an AtomGraph or similar structure).
        """
        self.graph1 = graph1
        self.graph2 = graph2

    @staticmethod
    def normalize(value, max_value):
        """
        Normalize a value between 0 and 1.

        **parameters:**
            - value: The value to normalize.
            - max_value: The maximum possible value for normalization.

        **returns:**
            - Normalized value between 0 and 1.
        """
        return min(value / (max_value + 1e-6), 1.0)

    @staticmethod
    def pad_targent(tensor, target_length):
        """
        Pad a tensor to the target length. Neccessary to
        improve the comparison between two features of the
        graphs.

        **parameters:**
            - tensor: The input tensor to pad or truncate.
            - target_length: The target length to match.

        **returns:**
            - A tensor of length equal to target_length.
        """
        current_length = tensor.shape[0]
        if current_length == target_length:
            return tensor
        elif current_length < target_length:
            padding = torch.zeros(
                (target_length - current_length,) + tensor.shape[1:], dtype=tensor.dtype)
            return torch.cat([tensor, padding], dim=0)
        else:
            return tensor[:target_length]

    def compute_similarity_index(self, weights=None):
        """
        Compute a similarity index between two AtomGraphs using a padding strategy to equalize sizes.

        **parameters:**
            - weights: Optional dictionary to assign different importance to various similarity metrics.

        **returns:**
            - A similarity index between 0 and 1.
        """

        if weights is None:
            weights = {
                'n_nodes_diff': 0.1,
                'n_edges_diff': 0.1,
                'atomic_numbers_diff': 0.4,
                'embedding_diff': 0.2,
                'senders_diff': 0.1,
                'receivers_diff': 0.1,
                'edge_vectors_diff': 0.1,
                'positions_diff': 0.1
            }

        # - Check number of nodes and edges
        n_nodes_diff = torch.abs(
            self.graph1.n_node - self.graph2.n_node).item()
        n_edges_diff = torch.abs(
            self.graph1.n_edge - self.graph2.n_edge).item()

        # - Compare atomic numbers
        atomic_numbers_1 = self.graph1.node_features['atomic_numbers']
        atomic_numbers_2 = self.graph2.node_features['atomic_numbers']

        max_len_atomic = max(len(atomic_numbers_1), len(atomic_numbers_2))
        atomic_numbers_1_padded = self.pad_targent(
            atomic_numbers_1, max_len_atomic)
        atomic_numbers_2_padded = self.pad_targent(
            atomic_numbers_2, max_len_atomic)
        atomic_diff = torch.sum(atomic_numbers_1_padded !=
                                atomic_numbers_2_padded).item() / max_len_atomic

        # - Compare node embeddings
        embeddings_1 = self.graph1.node_features['atomic_numbers_embedding']
        embeddings_2 = self.graph2.node_features['atomic_numbers_embedding']

        max_len_embed = max(embeddings_1.shape[0], embeddings_2.shape[0])
        embeddings_1_padded = self.pad_targent(embeddings_1, max_len_embed)
        embeddings_2_padded = self.pad_targent(embeddings_2, max_len_embed)
        embedding_diff = torch.norm(
            embeddings_1_padded - embeddings_2_padded, p='fro').item()

        # - Compare edge senders and receivers
        senders_1 = self.graph1.senders
        senders_2 = self.graph2.senders
        receivers_1 = self.graph1.receivers
        receivers_2 = self.graph2.receivers

        max_len_senders = max(len(senders_1), len(senders_2))
        senders_1_padded = self.pad_targent(senders_1, max_len_senders)
        senders_2_padded = self.pad_targent(senders_2, max_len_senders)
        senders_diff = torch.sum(
            senders_1_padded != senders_2_padded).item() / max_len_senders

        max_len_receivers = max(len(receivers_1), len(receivers_2))
        receivers_1_padded = self.pad_targent(
            receivers_1, max_len_receivers)
        receivers_2_padded = self.pad_targent(
            receivers_2, max_len_receivers)
        receivers_diff = torch.sum(
            receivers_1_padded != receivers_2_padded).item() / max_len_receivers

        # - Compare edge vectors
        edge_vectors_1 = self.graph1.edge_features['vectors']
        edge_vectors_2 = self.graph2.edge_features['vectors']

        max_len_edges = max(edge_vectors_1.shape[0], edge_vectors_2.shape[0])
        edge_vectors_1_padded = self.pad_targent(
            edge_vectors_1, max_len_edges)
        edge_vectors_2_padded = self.pad_targent(
            edge_vectors_2, max_len_edges)
        edge_diff = torch.norm(edge_vectors_1_padded -
                               edge_vectors_2_padded, p='fro').item()

        # - Compare positions (only if the number of nodes match)
        positions_1 = self.graph1.node_features['positions']
        positions_2 = self.graph2.node_features['positions']

        max_len_pos = max(positions_1.shape[0], positions_2.shape[0])
        positions_1_padded = self.pad_targent(positions_1, max_len_pos)
        positions_2_padded = self.pad_targent(positions_2, max_len_pos)
        position_diff = torch.norm(
            positions_1_padded - positions_2_padded, p='fro').item()

        # - Normalize the differences
        n_nodes_sim = 1 - self.normalize(n_nodes_diff, max_value=max(
            self.graph1.n_node.item(), self.graph2.n_node.item()))
        n_edges_sim = 1 - self.normalize(n_edges_diff, max_value=max(
            self.graph1.n_edge.item(), self.graph2.n_edge.item()))
        atomic_numbers_sim = 1 - self.normalize(atomic_diff, max_value=1.0)
        embedding_sim = 1 - \
            self.normalize(embedding_diff, max_value=embedding_diff + 1e-6)
        senders_sim = 1 - senders_diff
        receivers_sim = 1 - receivers_diff
        edge_vectors_sim = 1 - \
            self.normalize(edge_diff, max_value=edge_diff + 1e-6)
        position_sim = 1 - \
            self.normalize(position_diff, max_value=position_diff + 1e-6)

        # - Combine similarities into a single index
        similarity_index = (
            weights['n_nodes_diff'] * n_nodes_sim +
            weights['n_edges_diff'] * n_edges_sim +
            weights['atomic_numbers_diff'] * atomic_numbers_sim +
            weights['embedding_diff'] * embedding_sim +
            weights['senders_diff'] * senders_sim +
            weights['receivers_diff'] * receivers_sim +
            weights['edge_vectors_diff'] * edge_vectors_sim +
            weights['positions_diff'] * position_sim
        )

        return min(similarity_index, 1.0)



#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################
import pandas as pd
import argparse
import requests
import pubchempy as pcp


def opsin_name_to_smile(chemical_name):
    '''
    Function that uses opsin api to convert iupac names
    to cheminformatic identifiers

    **parameters**:
        chemical_name (str): The iupac name of the chemical.
    **returns**:
        smiles (str): The chemical structure in smiles format.
        inchi (str): The chemical structure in inchi format.
    '''
    opsin_url = 'https://opsin.ch.cam.ac.uk/opsin/'
    response = requests.get(opsin_url + chemical_name)
    print (response)
    output = response.json()
    if output['status'] == 'SUCCESS':
        return output
    else:
        return None


def pubchem_to_inchikey(identifier):
    '''
    Function that uses pubchem api to convert chemical names
    to inchi keys and inchi

    **parameters**:
        identifier (str): The name of the chemical.
    **returns**:
        inchikey (str): The inchi key of the chemical.
        smiles (str): The chemical structure in smiles format.
    '''
    pubchem = pcp.get_compounds(identifier, 'name')
    if len(pubchem) > 0:
        return pubchem[0].inchikey, pubchem[0].canonical_smiles
    else:
        return None


def name_to_cheminfo(chemical_name):
    """
    Converts a chemical name into cheminformatic identifiers
    using OPSIN and PubChem APIs.

    This function attempts to retrieve the InChIKey and SMILES
    representation of a
    given chemical name. It first tries to use the OPSIN API to fetch the data.
    If OPSIN does not return a result, it falls back to using the PubChem API.

    Parameters:
        chemical_name (str): The name of the chemical to be converted.
                             Can be an IUPAC name or other
                             recognized chemical name.

    Returns:
        dict: A dictionary containing the following keys:
              - 'inchikey': The InChIKey of the chemical (str),
              or None if not found.
              - 'smiles': The SMILES representation
              of the chemical (str), or None if not found.

              If no information is found from either API,
              returns an empty dictionary.
    """
    tmp = {}
    data = opsin_name_to_smile(chemical_name)
    if data != None:
        tmp['inchikey'] = data['stdinchikey']
        tmp['smiles'] = data['smiles']
    else:
        data = pubchem_to_inchikey(chemical_name)
        if data != None:

            tmp['inchikey'] = data[0]
            tmp['smiles'] = data[1]
    if len(tmp) > 0:
        return {'name': chemical_name, **tmp}
    else:
        return tmp


def print_helpful_information():
    '''
    Prints helpful information about using the chemical_parser script.
    '''
    help_text = """
    Usage: chemical_parser.py [CHEMICAL_NAME] [OPTIONS]

    This script takes a chemical name as input, retrieves cheminformatic identifiers
    (InChIKey and SMILES), and writes the information to a CSV file.

    Positional Arguments:
      CHEMICAL_NAME       The name of the chemical to parse.

    Optional Arguments:
      -o, --output FILE   The path to the output CSV file (default: cheminfor).
      -n, --name NAME     The name of the chemical to parse (alternative to positional argument).

    Examples:
      python chemical_parser.py "water" -o water_info.csv
      python chemical_parser.py -n "ethanol"
    """
    print(help_text)

def main():
    parser = argparse.ArgumentParser(description="Parse a chemical name and write cheminformatic identifiers to a CSV file.")
    parser.add_argument('chemical_name', type=str, nargs='?', default=None, help='The name of the chemical to parse.')
    parser.add_argument('-n', '--name', type=str, help='The name of the chemical to parse (alternative to positional argument).')
    parser.add_argument('-o', '--output', type=str, default='cheminfor', help='The path to the output CSV file (default: output.csv).')

    args = parser.parse_args()

    chemical_name = args.name if args.name else args.chemical_name
    output_csv = args.output

    if chemical_name is None:
        print_helpful_information()
        return


    chemical_info = name_to_cheminfo(chemical_name)

    if chemical_info:
        df = pd.DataFrame([chemical_info])
        df.to_csv(f'{output_csv}.csv', index=False)
        print(f"Chemical information written to {output_csv}.csv")
    else:
        print("No chemical information found.")

if __name__ == '__main__':
    main()




#!/usr/bin/python
from __future__ import print_function
__author__ = "Dr. Dinga Wonanke"
__status__ = "production"

##############################################################################
# fairmofsyncondition is a machine learning package for predicting the        #
# synthesis condition of the crystal structures of MOFs. It is also intended  #
# for predicting all MOFs the can be generated from a given set of conditions #
# In addition the package also predicts the stability of MOFs, compute their  #
# their PXRD and crystallite sizes. This package is part of our effort to     #
# to accelerate the discovery and optimization of the synthesises of novel    #
# high performing MOFs. This package is being developed by Dr Dinga Wonanke   #
# as part of hos MSCA post doctoral fellowship at TU Dresden.                 #
#                                                                             #
###############################################################################

import os
import sys
import torch
import random
import argparse
import inspect
from torch import nn
from torch.nn import functional
from torch import optim
from tqdm import tqdm
from tabulate import tabulate
from torch.utils.tensorboard import SummaryWriter
from torch_geometric.nn import GATv2Conv, global_mean_pool, GINConv, BatchNorm, global_max_pool, global_add_pool
from torch_geometric.loader import DataLoader
from fairmofsyncondition.read_write import coords_library, filetyper

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class EnergyGNN_GAT2Conv(nn.Module):
    """
    A Graph Neural Network class for predicting the thermodynamic
    stability of MOFs using Graph Attention Networks (GATv2).

    Arg:
        input_dim (int): Number of input node features.
        hidden_dim (int): Number of hidden units in the GATv2 layers.
        output_dim (int): Number of output units (e.g., 1 for regression).
        heads (int, optional): Number of attention heads. Default is 1.
        dropout (float, optional): Dropout rate. Default is 0.2.
    """

    def __init__(self,
                 input_dim,
                 hidden_dim,
                 output_dim,
                 edge_dim,
                 heads=4,
                 dropout=0.2
                 ):
        super(EnergyGNN_GAT2Conv, self).__init__()
        input_dim = int(input_dim)
        hidden_dim = int(hidden_dim)
        output_dim = int(output_dim)
        edge_dim = int(edge_dim)
        heads = int(heads)
        self.conv1 = GATv2Conv(input_dim, hidden_dim,
                               heads=heads, edge_dim=edge_dim)
        self.norm1 = nn.BatchNorm1d(hidden_dim * heads)
        self.conv2 = GATv2Conv(
            hidden_dim * heads, hidden_dim, heads=heads, edge_dim=edge_dim)
        self.norm2 = nn.BatchNorm1d(hidden_dim * heads)
        self.conv3 = GATv2Conv(
            hidden_dim * heads, hidden_dim, heads=heads, edge_dim=edge_dim)
        self.norm3 = nn.BatchNorm1d(hidden_dim * heads)
        self.fc1 = nn.Linear(hidden_dim * heads, hidden_dim)
        self.dropout = nn.Dropout(p=dropout)
        self.fc2 = nn.Linear(hidden_dim, output_dim)

    def forward(self, data):
        x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch
        x = functional.leaky_relu(self.norm1(
            self.conv1(x, edge_index, edge_attr)))
        x = functional.leaky_relu(self.norm2(
            self.conv2(x, edge_index, edge_attr)))
        x = functional.leaky_relu(self.norm3(
            self.conv3(x, edge_index, edge_attr)))
        x = global_mean_pool(x, batch)
        x = functional.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        return x


class EnergyGNN_GIN(nn.Module):
    """
    Implements a Graph Neural Network (GNN) using Graph Isomorphism Network (GIN) layers
    for graph-level prediction tasks such as regression or classification.

    Attributes:
        input_dim (int): Dimensionality of input node features.
        hidden_dim (int): Dimensionality of hidden layers.
        output_dim (int): Dimensionality of the output.
        dropout (float): Dropout rate for regularization. Default is 0.2.

    Layers:
        - Three GINConv layers, each using an MLP for message passing.
        - Batch normalization after each GINConv layer.
        - Global mean pooling for aggregating node-level features to graph-level.
        - Fully connected layers for output prediction.

    Methods:
        forward(data):
            Performs a forward pass of the model on input graph data.
    """

    def __init__(self,
                 input_dim,
                 hidden_dim,
                 output_dim,
                 edge_dim=None,
                 heads=None,
                 dropout=0.2
                 ):
        """
        Initializes the GNN model.

        Args:
            input_dim (int): Dimensionality of input node features.
            hidden_dim (int): Dimensionality of hidden layers.
            output_dim (int): Dimensionality of the output.
            edge_dim (int, optional): Dimensionality of edge features. Not used in this implementation.
            heads (int, optional): Number of attention heads. Not used in this implementation.
            dropout (float, optional): Dropout rate for regularization. Default is 0.2.
        """
        super(EnergyGNN_GIN, self).__init__()
        input_dim = int(input_dim)
        hidden_dim = int(hidden_dim)
        output_dim = int(output_dim)

        # Define MLPs for GINConv layers
        self.mlp1 = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim)
        )
        self.conv1 = GINConv(self.mlp1)

        self.mlp2 = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim)
        )
        self.conv2 = GINConv(self.mlp2)

        self.mlp3 = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim)
        )
        self.conv3 = GINConv(self.mlp3)

        # Batch normalization layers
        self.norm1 = BatchNorm(hidden_dim)
        self.norm2 = BatchNorm(hidden_dim)
        self.norm3 = BatchNorm(hidden_dim)

        # Fully connected layers
        self.fc1 = nn.Linear(hidden_dim, hidden_dim)
        self.dropout = nn.Dropout(p=dropout)
        self.fc2 = nn.Linear(hidden_dim, output_dim)

    def forward(self, data):
        """
        Defines the forward pass for the model.

        Args:
            data (torch_geometric.data.Data): Input graph data object containing:
                - x (torch.Tensor): Node features of shape [num_nodes, input_dim].
                - edge_index (torch.Tensor): Edge indices of shape [2, num_edges].
                - edge_attr (torch.Tensor, optional): Edge features (not used here).
                - batch (torch.Tensor): Batch indices for mini-batch training.

        Returns:
            torch.Tensor: Final output predictions of shape [num_graphs, output_dim].
        """
        x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch

        # Apply GIN layers with normalization and ReLU activations
        x = functional.relu(self.norm1(self.conv1(x, edge_index)))
        x = self.dropout(x)
        x = functional.relu(self.norm2(self.conv2(x, edge_index)))
        x = self.dropout(x)
        x = functional.relu(self.norm3(self.conv3(x, edge_index)))
        x = self.dropout(x)

        # Global mean pooling
        x = global_mean_pool(x, batch)  # Mean pooling

        # Fully connected layers with dropout
        x = functional.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.fc2(x)
        return functional.relu(x)


class GraphLatticeModel(nn.Module):
    """
    A Graph Neural Network (GNN) model that incorporates both graph-based features and lattice matrix features
    for graph-level prediction tasks such as property regression or classification.

    Attributes:
        input_dim (int): Dimensionality of input node features.
        gnn_hidden_dim (int): Dimensionality of GNN hidden layers.
        lattice_hidden_dim (int): Dimensionality of hidden layers for lattice matrix processing.
        output_dim (int): Dimensionality of the output (e.g., target properties).
        num_layers (int): Number of GNN layers. Default is 3.
        dropout (float): Dropout rate for regularization. Default is 0.2.

    Methods:
        forward(data):
            Performs a forward pass through the model using the input graph and lattice data.
    """

    def __init__(self,
                 input_dim,
                 gnn_hidden_dim,
                 lattice_hidden_dim,
                 output_dim,
                 num_layers=3,
                 dropout=0.2
                 ):
        """
        Initializes the GraphLatticeModel.

        Args:
            input_dim (int): Dimensionality of input node features.
            gnn_hidden_dim (int): Dimensionality of GNN hidden layers.
            lattice_hidden_dim (int): Dimensionality of hidden layers for lattice matrix processing.
            output_dim (int): Dimensionality of the output.
            num_layers (int, optional): Number of GNN layers. Default is 3.
            dropout (float, optional): Dropout rate for regularization. Default is 0.2.
        """
        super(GraphLatticeModel, self).__init__()
        input_dim = int(input_dim)
        hidden_dim = int(gnn_hidden_dim)
        lattice_hidden_dim = int(lattice_hidden_dim)
        output_dim = int(output_dim)
        num_layers = int(num_layers)

        # Define GINConv layers, their MLPs, and batch normalization dynamically
        self.convs = nn.ModuleList()
        self.norms = nn.ModuleList()
        for i in range(num_layers):
            mlp = nn.Sequential(
                nn.Linear(input_dim if i == 0 else hidden_dim, hidden_dim),
                nn.ReLU(),
                nn.Linear(hidden_dim, hidden_dim)
            )
            self.convs.append(GINConv(mlp))
            self.norms.append(BatchNorm(hidden_dim))

        # MLP for lattice matrix processing
        self.mlp = nn.Sequential(
            # Lattice matrix flattened (3x3 = 9)
            nn.Linear(9, lattice_hidden_dim),
            nn.ReLU(),
            nn.Linear(lattice_hidden_dim, lattice_hidden_dim),
            nn.ReLU()
        )

        # Fully connected layers for final prediction
        combined_dim = gnn_hidden_dim + lattice_hidden_dim
        self.fc1 = nn.Linear(combined_dim, gnn_hidden_dim)
        self.fc2 = nn.Linear(gnn_hidden_dim, output_dim)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, data):
        """
        Performs a forward pass of the model.

        Args:
            data (torch_geometric.data.Data): Input data containing:
                - x (torch.Tensor): Node features of shape [num_nodes, input_dim].
                - edge_index (torch.Tensor): Edge indices of shape [2, num_edges].
                - edge_attr (torch.Tensor, optional): Edge features (not used here).
                - batch (torch.Tensor): Batch indices for mini-batch training.
                - lattice (torch.Tensor): Lattice matrix of shape [num_graphs, 3, 3].

        Returns:
            torch.Tensor: Final predictions of shape [num_graphs, output_dim].
        """
        x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch
        lattice = data.lattice.view(data.lattice.size(0), -1)

        # Apply GIN layers with normalization and ReLU activations
        for conv, norm in zip(self.convs, self.norms):
            x = functional.relu(norm(conv(x, edge_index)))
            x = self.dropout(x)

        # Global mean pooling
        x = global_mean_pool(x, batch)  # Mean pooling

        # MLP processing of lattice matrix
        lattice_rep = self.mlp(lattice)

        # Concatenate graph and lattice representations
        combined_rep = torch.cat([x, lattice_rep], dim=1)

        # Fully connected layers with dropout
        combined_rep = functional.relu(self.fc1(combined_rep))
        combined_rep = self.dropout(combined_rep)
        out = self.fc2(combined_rep)

        # Scale output to [0, 1]
        out = torch.tanh(out) * 0.5 + 0.5
        return out


class EarlyStopping:
    """
    Implements early stopping to terminate training when validation loss stops improving.

    Attributes:
        patience (int): Number of epochs/iteractions to wait for an improvement in validation loss before stopping. Default is 5.
        delta (float): Minimum improvement in validation loss required to reset the patience counter. Default is 0.001.
        best_loss (float): The lowest validation loss observed so far. Initialized to infinity.
        counter (int): Tracks the number of consecutive epochs without improvement in validation loss.
        early_stop (bool): Flag indicating whether early stopping condition is met.
        path (str): File path to save the best model checkpoint. Default is 'checkpoint.pth'.

    Methods:
        __call__(val_loss, model):
            Evaluates the validation loss and decides whether to stop training or save a checkpoint.
        save_checkpoint(model):
            Saves the model's state dictionary to the specified checkpoint file.
    """

    def __init__(self,
                 patience=5,
                 delta=0.001,
                 path='checkpoint.pth'
                 ):
        """
        Initializes the EarlyStopping object with specified parameters.

        Args:
            patience (int, optional): Number of epochs to wait for an improvement in validation loss. Default is 5.
            delta (float, optional): Minimum improvement in validation loss required. Default is 0.001.
            path (str, optional): Path to save the best model checkpoint. Default is 'checkpoint.pth'.
        """
        self.patience = patience
        self.delta = delta
        self.best_loss = float('inf')
        self.counter = 0
        self.early_stop = False
        self.path = path

    def __call__(self,
                 val_loss,
                 model,
                 optimizer,
                 normalise_parameter
                 ):
        """
        Evaluates the validation loss and decides whether to stop training or save a model checkpoint.

        Args:
            val_loss (float): The current validation loss.
            model (torch.nn.Module): The PyTorch model being trained.

        Behavior:
            - If the validation loss improves by at least `delta`, resets the counter and saves a checkpoint.
            - Otherwise, increments the counter.
            - Stops training if the counter exceeds or equals `patience`.
        """
        if val_loss < self.best_loss - self.delta:
            self.best_loss = val_loss
            self.counter = 0
            self.save_checkpoint(model, optimizer, normalise_parameter)
        else:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True


    def save_checkpoint(self,
                        model,
                        optimizer,
                        normalise_parameter
                        ):
        """
        Saves the current model state to the checkpoint file.

        Args:
            model (torch.nn.Module): The PyTorch model being trained.
            optimizer (optim.Optimizer): Optimizer for updating model parameters.
            normalise_parameter (float): Normalization parameter for target values.

        Behavior:
            - Saves the model's state dictionary to the file specified by `path`.
        """
        save_model(model, optimizer, normalise_parameter, self.path)


def train(model,
          dataloader,
          optimizer,
          criterion,
          device
          ):
    """
    Train the model using the given data and optimizer.

    **parameters:**
        model (nn.Module): The GNN model to train
        dataloader (DataLoader): DataLoader for batching the dataset during training.
        optimizer (optim.Optimizer): Optimizer for updating model parameters.
        criterion (nn.Module): Loss function (e.g., MSELoss) to compute the training loss.
        device (torch.device): The device (CPU or GPU) for computation.

    **returns:**
        float: The average training loss over the epoch.
    """

    model.train()
    total_loss = 0
    for data in dataloader:
        data.lattice = data.lattice.reshape(-1,3,3)
        data = data.to(device)
        optimizer.zero_grad()
        predictions = model(data).view(-1)
        loss = criterion(predictions, data.y)
        loss.backward()
        optimizer.step()
        total_loss += loss.item()
    return total_loss / len(dataloader)


def normalize_data(dataset, method='z-score'):
    """
    Normalize the target values (data.y) in the dataset.

    Parameters:
        dataset (Dataset): The dataset object containing the data.
        method (str): The normalization method ('min-max' or 'z-score').

    Returns:
        tuple: Normalized dataset and the normalization parameters.
    """
    y_values = torch.cat([data.y for data in dataset]
                         )  # Gather all target values
    if method == 'min-max':
        y_min, y_max = y_values.min(), y_values.max()
        for data in dataset:
            data.y = (data.y - y_min) / (y_max - y_min)
        return dataset, {'min': y_min, 'max': y_max}

    elif method == 'z-score':
        y_mean, y_std = y_values.mean(), y_values.std()
        for data in dataset:
            data.y = (data.y - y_mean) / y_std
        return dataset, {'mean': y_mean, 'std': y_std}

    else:
        raise ValueError(
            "Unsupported normalization method. Choose 'min-max' or 'z-score'.")


def inverse_normalize(predictions,
                      normalization_params,
                      method='z-score'):
    """
    Inverse the normalization of predictions to the original scale.

    Parameters:
        predictions (Tensor): The normalized predictions.
        normalization_params (dict): The parameters used for normalization.
        method (str): The normalization method ('min-max' or 'z-score').

    Returns:
        Tensor: Predictions in the original scale.
    """
    if method == 'min-max':
        return predictions * (normalization_params['max'] - normalization_params['min']) + normalization_params['min']
    elif method == 'z-score':
        return predictions * normalization_params['std'] + normalization_params['mean']
    else:
        raise ValueError(
            "Unsupported normalization method. Choose 'min-max' or z-score")


def evaluate(model,
             dataloader,
             criterion,
             device
             ):
    """
    Evaluate the model using the given data and loss function, and compute accuracy.

    **parameters:**
        model (nn.Module): The trained model to evaluate.
        dataloader (DataLoader): DataLoader for batching the dataset during evaluation.
        criterion (nn.Module): Loss function (e.g., CrossEntropyLoss) to compute the evaluation loss.
        device (torch.device): The device (CPU or GPU) for computation.

    **returns:**
        tuple: A tuple containing the average evaluation loss and accuracy.
               (average_loss, accuracy)
    """

    model.eval()
    total_loss = 0
    correct_predictions = 0
    total_samples = 0

    with torch.no_grad():
        for data in dataloader:
            data.lattice = data.lattice.reshape(-1,3,3)
            data = data.to(device)
            labels = data.y
            predictions = model(data).view(-1)
            loss = criterion(predictions, labels)
            total_loss += loss.item()
            total_samples += labels.size(0)

    average_loss = total_loss / len(dataloader)

    return average_loss


def save_model(model,
               optimizer,
               normalise_parameter,
               path="model.pth"
               ):
    """
    Save the trained model and optimizer state to a file.

    **parameters:**
        model (nn.Module): The trained GNN model to save.
        optimizer (optim.Optimizer): Optimizer for updating model parameters.
        path (str, optional): Path to save the model. Default is "model.pth".
    """
    checkpoint = {
        'model_state_dict': model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict(),
        'model_architecture': model.__class__,
        'model_args': model.__dict__.get('_modules'),
        'normalise_parameter': normalise_parameter
    }
    torch.save(checkpoint, path)


def load_model(path="model.pth", device="cpu"):
    """
    Load a saved model and optimizer state from a file.

    **parameters:**
        path (str, optional): Path to load the model. Default is "model.pth".
        device (torch.device, optional): The device (CPU or GPU) for computation. Default is "cpu".

    **returns:**
        tuple: The loaded model and optimizer.
    """
    checkpoint = torch.load(path, map_location=device)
    model_class = checkpoint['model_architecture']
    model_args = checkpoint['model_args']
    normalise_parameter = checkpoint['normalise_parameter']
    model = model_class(**model_args).to(device)
    model.load_state_dict(checkpoint['model_state_dict'])

    optimizer = optim.Adam(model.parameters())  # Rebuild the optimizer
    optimizer.load_state_dict(checkpoint['optimizer_state_dict'])

    return model, optimizer, normalise_parameter


def transform_target(test_data,
                     normalize_param,
                     method='z-score'
                     ):
    """
    Transform the target values in the test dataset according to the normalization parameters.

    Parameters:
        test_data (Dataset): The test dataset containing the target values.
        normalize_param (dict): The normalization parameters (mean and std).
        method (str, optional): The normalization method ('z-score' or'min-max'). Default is 'z-score'.

    Returns:
        Dataset: The transformed test dataset.
    """
    if method == 'z-score':
        mean, std = normalize_param['mean'], normalize_param['std']
        for data in test_data:
            data.y = (data.y - mean) / std
    elif method == 'min-max':
        min_val, max_val = normalize_param['min'], normalize_param['max']
        for data in test_data:
            data.y = (data.y - min_val) / (max_val - min_val)
    else:
        raise ValueError(
            "Unsupported normalization method. Choose 'z-score' or'min-max'.")
    return test_data


def load_dataset(path_to_lmdb,
                 batch_size,
                 train_size=0.9,
                 random_seed=42,
                 shuffle=True,
                 normalize='full'
                 ):
    """
    Loads a dataset from an LMDB file and splits it into training, validation, and test sets.

    The function uses the `coords_library.LMDBDataset` to load the dataset and splits it into
    training and test datasets. The training dataset is further split into training and validation
    sets. Data loaders are created for the training and validation datasets.

    Parameters:
        path_to_lmdb (str):
            Path to the LMDB file containing the dataset.
        batch_size (int):
            Batch size for the DataLoader.
        train_size (float, optional):
            Fraction of the data to use for training. The rest is used for testing. Default is 0.8.
        random_seed (int, optional):
            Random seed for splitting the data. Ensures reproducibility. Default is 42.
        shuffle (bool, optional):
            Whether to shuffle the data before splitting. Default is True.
        normalize (str, optional):
            Normalization method to use. Can be 'full' for full normalization or 'batch' for
            batch normalization. Default is 'full'.

    Returns:
        tuple:
            - train_loader (DataLoader): DataLoader for the training dataset.
            - val_loader (DataLoader): DataLoader for the validation dataset.
            - test_dataset (Dataset): Dataset object containing the test data.
    """
    dataset = coords_library.LMDBDataset(path_to_lmdb)
    train_dataset, test_dataset = dataset.split_data(
        train_size=train_size, random_seed=random_seed, shuffle=shuffle)

    train_indices, val_indices = coords_library.list_train_test_split(list(range(len(train_dataset))))
    train_data = train_dataset[train_indices]
    val_data = train_dataset[val_indices]
    if normalize == 'full':
        train_data_norm, normalise_parameter = normalize_data(train_data)
        train_loader = DataLoader(train_data_norm, batch_size=batch_size, shuffle=True)
        val_data = transform_target(val_data, normalise_parameter)
        test_dataset = transform_target(test_dataset, normalise_parameter)
    elif normalize == 'batch':
        print("error")
        train_loader = DataLoader(train_data_norm, batch_size=batch_size, shuffle=True)
        train_loader, normalise_parameter = normalize_data(train_loader)
        val_data = transform_target(val_data, normalise_parameter)
        test_dataset = transform_target(test_dataset, normalise_parameter)

    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    val_loader = DataLoader(val_data, batch_size=batch_size, shuffle=False)
    return train_loader, val_loader, test_dataset, normalise_parameter


def GAT2Conv_main_model(path_to_lmdb,
                        hidden_dim,
                        learning_rate,
                        batch_size,
                        dropout,
                        heads,
                        epoch,
                        patience,
                        print_every,
                        save_path
                        ):
    """
    Main function for training the GAT2Conv model.

    **parameters**:
        path_to_lmdb (str): Path to the LMDB file containing the dataset.
        hidden_dim (int): Dimensionality of the hidden layer.
        learning_rate (float): Learning rate for the optimizer.
        batch_size (int): Batch size for training.
        dropout (float): Dropout rate for the model.
        heads (int): Number of attention heads for the GAT layer.
        epoch (int): Number of training epochs.
        patience (int): Patience for early stopping.
        print_every (int): Frequency of printing training loss and validation loss.
        save_path (str): Path to save the best model.

    **Bahaviors **
        This function loads the dataset, trains the GAT2Conv model, and saves the best model.
    """

    writer = SummaryWriter(log_dir="errorlogger/energy_gat")
    early_stopping = EarlyStopping(
        patience=patience, path= 'best_'+save_path)
    model = EnergyGNN_GAT2Conv(input_dim=4,
                               hidden_dim=hidden_dim,
                               output_dim=1,
                               edge_dim=1,
                               heads=heads,
                               dropout=dropout).to(device)

    optimizer = optim.Adam(model.parameters(), lr=learning_rate, weight_decay=0.00001)

    criterion = nn.L1Loss()

    train_loader, val_loader, test_dataset, normalise_parameter = load_dataset(
        path_to_lmdb, batch_size)
    for i in tqdm(range(epoch)):
        train_loss = train(model, train_loader, optimizer, criterion, device)
        val_loss = evaluate(model, val_loader, criterion, device)
        early_stopping(val_loss, model, optimizer, normalise_parameter)
        if early_stopping.early_stop:
            print("Early stopping triggered.")
            break
        if i % print_every == 0:
            print(
                f"Epoch: {i+1}/{epoch}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}")

    # model.load_state_dict(torch.load(save_path))
    # model, optimizer, normalise_parameter = load_model(save_path)
    save_model(model, optimizer, normalise_parameter, save_path)

    val_loss = evaluate(model, val_loader, criterion, device)
    print("Loaded the best model saved during training.", val_loss)


def EnergyGNN_GIN_main_model(path_to_lmdb,
                             hidden_dim,
                             learning_rate,
                             batch_size,
                             dropout,
                             epoch,
                             patience,
                             print_every,
                             save_path):
    """
    Main function for training the EnergyGNN_GIN model.

    **parameters**:
        path_to_lmdb (str): Path to the LMDB file containing the dataset.
        hidden_dim (int): Dimensionality of the hidden layer.
        learning_rate (float): Learning rate for the optimizer.
        batch_size (int): Batch size for training.
        dropout (float): Dropout rate for the model.
        epoch (int): Number of training epochs.
        patience (int): Patience for early stopping.
        print_every (int): Frequency of printing training loss and validation loss.
        save_path (str): Path to save the best model.

    **Bahaviors **
        This function loads the dataset, trains the EnergyGNN_GIN model, and saves the best model.
    """


    writer = SummaryWriter(log_dir="errorlogger/energy_gat")
    early_stopping = EarlyStopping(patience=patience, path=save_path)
    model = EnergyGNN_GIN(input_dim=4,
                          hidden_dim=hidden_dim,
                          output_dim=1,
                          edge_dim=1,
                          dropout=dropout).to(device)
    optimizer = optim.Adam(
        model.parameters(), lr=learning_rate, weight_decay=0.00001)
    criterion = nn.L1Loss()

    train_loader, val_loader, test_dataset, normalise_parameter = load_dataset(
        path_to_lmdb, batch_size)

    for i in tqdm(range(epoch)):
        train_loss = train(model, train_loader, optimizer, criterion, device)
        val_loss = evaluate(model, val_loader, criterion, device)
        early_stopping(val_loss, model, optimizer, normalise_parameter)

        if early_stopping.early_stop:
            print("Early stopping triggered.")
            break
        if i % print_every == 0:
            print(
                f"Epoch: {i+1}/{epoch}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}")

    save_model(model, optimizer, normalise_parameter, save_path)

    val_loss = evaluate(model, val_loader, criterion, device)
    print("Loaded the best model saved during training.", val_loss)


def GraphLatticeModel_main_model(path_to_lmdb,
                                 hidden_dim,
                                 lattice_hidden_dim,
                                 num_layers,
                                 learning_rate,
                                 batch_size,
                                 dropout,
                                 epoch,
                                 patience,
                                 print_every,
                                 save_path
                                 ):
    """
    Main function for training the EnergyGNN_GIN model.

    **parameters**:
        path_to_lmdb (str): Path to the LMDB file containing the dataset.
        hidden_dim (int): Dimensionality of the hidden layer.
        learning_rate (float): Learning rate for the optimizer.
        batch_size (int): Batch size for training.
        dropout (float): Dropout rate for the model.
        epoch (int): Number of training epochs.
        patience (int): Patience for early stopping.
        print_every (int): Frequency of printing training loss and validation loss.
        save_path (str): Path to save the best model.

    **Bahaviors **
        This function loads the dataset, trains the EnergyGNN_GIN model, and saves the best model.
    """

    writer = SummaryWriter(log_dir="errorlogger/gat2conv")
    early_stopping = EarlyStopping(patience=patience, path=save_path)

    model = GraphLatticeModel(input_dim=4,
                              gnn_hidden_dim=hidden_dim,
                              lattice_hidden_dim=lattice_hidden_dim,
                              output_dim=1,
                              num_layers=num_layers,
                              dropout=dropout).to(device)
    optimizer = optim.Adam(
        model.parameters(), lr=learning_rate, weight_decay=0.00001)
    criterion = nn.L1Loss()

    train_loader, val_loader, test_dataset, normalise_parameter = load_dataset(
        path_to_lmdb, batch_size)

    for i in tqdm(range(epoch)):
        train_loss = train(model, train_loader, optimizer, criterion, device)
        val_loss = evaluate(model, val_loader, criterion, device)
        early_stopping(val_loss, model, optimizer, normalise_parameter)

        if early_stopping.early_stop:
            print("Early stopping triggered.")
            break
        if i % print_every == 0:
            print(
                f"Epoch: {i+1}/{epoch}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}")

    save_model(model, optimizer, normalise_parameter, save_path)

    val_loss = evaluate(model, val_loader, criterion, device)
    print("Loaded the best model saved during training.", val_loss)


def display_model_helper(model):
    """
    Display clear and concise information about the selected model and its key parameters.

    Args:
        model (str): The name of the model. Options are:
                     - "EnergyGNN_GAT2Conv": Graph Attention Networks (GATv2) based model.
                     - "EnergyGNN_GIN": Graph Isomorphism Network (GIN) based model.
                     - "GraphLatticeModel": Combines graph features with lattice matrix processing.

    Instructions:
        - Review the description for each model.
        - Identify the required parameters for your use case.
        - Ensure you provide the necessary inputs when configuring the model.

    Example Usage:
        display_model_helper("EnergyGNN_GIN")
    """
    model_details = {
        "EnergyGNN_GAT2Conv": """
        Model: EnergyGNN_GAT2Conv
        -------------------------
        - Type: Graph Neural Network using GATv2 (Graph Attention Networks).
        - Suitable for: Models that require attention mechanisms on graph edges.
        - Key Parameters:
            - `hidden_dim`: Number of hidden units in the network layers (e.g., 128 or 256).
            - `heads`: Number of attention heads for the GAT layer (e.g. 2, 4 ,6 , 8).
            - `dropout`: Regularization rate to prevent overfitting (e.g., 0.2 - 0.9).
        """,
        "EnergyGNN_GIN": """
        Model: EnergyGNN_GIN
        --------------------
        - Type: Graph Neural Network using GIN (Graph Isomorphism Networks).
        - Suitable for: Dense and message-passing networks for graph regression tasks.
        - Key Parameters:
            - `hidden_dim`: Number of hidden units in the network layers (e.g., 128 or 256).
            - `dropout`: Regularization rate to prevent overfitting (e.g., 0.1 or 0.3).
        """,
        "GraphLatticeModel": """
        Model: GraphLatticeModel
        ------------------------
        - Type: Hybrid Graph and Lattice Neural Network.
        - Combines: Graph-based features and 3x3 lattice matrix for property predictions.
        - Suitable for: Tasks requiring structural and lattice-based properties.
        - Key Parameters:
            - `gnn_hidden_dim`: Number of hidden units for the graph network (e.g., 128 or 256).
            - `lattice_hidden_dim`: Number of hidden units for lattice matrix processing (e.g., 64 or 128).
            - `num_layers`: Number of GNN layers to use (e.g., 3 or 5).
            - `dropout`: Regularization rate to prevent overfitting (e.g., 0.2 or 0.5).
        """
    }

    if model in model_details:
        print(model_details[model])
    else:
        print(f"No helper available for the model '{model}'. Please choose from:")
        print("- EnergyGNN_GAT2Conv")
        print("- EnergyGNN_GIN")
        print("- GraphLatticeModel")

def parse_arguments():
    """
    Parse command-line arguments for training the GNN model.

    Returns:
        argparse.Namespace: Parsed command-line arguments with detailed descriptions.
    """
    parser = argparse.ArgumentParser(
        description=(
            "Train a Graph Neural Network (GNN) model for training a bond dissociation enthalpy predictor of Metal-Organic Frameworks (MOFs).\n"
            "Select the appropriate model and provide necessary hyperparameters.\n\n"
        ),
        formatter_class=argparse.RawTextHelpFormatter,  # Automatically show defaults in help
    )

    # General arguments
    parser.add_argument(
        '-p', '--path_to_lmdb',
        type=str,
        required=True,
        help="Path to the LMDB file containing the dataset.\n\n"
    )
    parser.add_argument(
        '-m', '--model',
        type=str,
        choices=["EnergyGNN_GAT2Conv", "EnergyGNN_GIN", "GraphLatticeModel"],
        required=True,
        help=(
            "Model to train. Choose one of the following:\n\n"
            "  - EnergyGNN_GAT2Conv: Graph Attention Network (GATv2).\n\n"
            "  - EnergyGNN_GIN: Graph Isomorphism Network (GIN).\n\n"
            "  - GraphLatticeModel: Combines graph features with lattice matrix.\n\n\n"
        )
    )

    # Model hyperparameters
    parser.add_argument(
        '-hd', '--hidden_dim',
        type=int,
        default=256,
        help="Number of hidden units in the network layers.\n\n"
    )
    parser.add_argument(
        '-lhd', '--lattice_hidden_dim',
        type=int,
        default=3,
        help=(
            "Number of hidden units for processing lattice matrix (only for GraphLatticeModel).\n\n"
        )
    )
    parser.add_argument(
        '-nl', '--num_layers',
        type=int,
        default=3,
        help=(
            "Number of GNN layers (used only in GraphLatticeModel).\n"
            "Ignored for other models.\n\n"
        )
    )

    # Training hyperparameters
    parser.add_argument(
        '-lr', '--learning_rate',
        type=float,
        default=0.001,
        help="Learning rate for training the model.\n\n"
    )
    parser.add_argument(
        '-bs', '--batch_size',
        type=int,
        default=2048,
        help="Batch size for training.\n\n"
    )
    parser.add_argument(
        '-d', '--dropout',
        type=float,
        default=0.1,
        help="Dropout rate for regularization to prevent overfitting.\n\n"
    )
    parser.add_argument(
        '-head', '--heads',
        type=int,
        default=4,
        help="Number of attention heads (only for GATv2-based models).\n\n"
    )
    parser.add_argument(
        '-patience', '--patience',
        type=int,
        default=200,
        help="Number of epochs to wait without improvement before early stopping.\n\n"
    )
    parser.add_argument(
        '-print_every',
        type=int,
        default=10,
        help="Frequency (in epochs) of printing training progress and validation loss.\n\n"
    )
    parser.add_argument(
        '-e', '--epoch',
        type=int,
        default=500,
        help="Total number of epochs for training.\n\n"
    )

    # Output options
    parser.add_argument(
        '-s', '--save_path',
        type=str,
        default="model_checkpoint",
        help="Path to save the best model checkpoint.\n\n"
    )

    return parser



def entry_point():
    """
    Entry point for the training script. Dynamically calls the appropriate main function based on the selected model.
    """
    parser = parse_arguments()
    args = parser.parse_args()

    # Display helper for the selected model
    display_model_helper(args.model)

    # Dynamically select the appropriate main function for the chosen model
    if args.model == "EnergyGNN_GAT2Conv":
        GAT2Conv_main_model(
            path_to_lmdb=args.path_to_lmdb,
            hidden_dim=args.hidden_dim,
            learning_rate=args.learning_rate,
            batch_size=args.batch_size,
            dropout=args.dropout,
            heads=args.heads,
            epoch=args.epoch,
            patience=args.patience,
            print_every=args.print_every,
            save_path=args.save_path,
        )
    elif args.model == "EnergyGNN_GIN":
        EnergyGNN_GIN_main_model(
            path_to_lmdb=args.path_to_lmdb,
            hidden_dim=args.hidden_dim,
            learning_rate=args.learning_rate,
            batch_size=args.batch_size,
            dropout=args.dropout,
            epoch=args.epoch,
            patience=args.patience,
            print_every=args.print_every,
            save_path=args.save_path,
        )
    elif args.model == "GraphLatticeModel":
        GraphLatticeModel_main_model(
            path_to_lmdb=args.path_to_lmdb,
            hidden_dim=args.hidden_dim,
            lattice_hidden_dim=args.lattice_hidden_dim,
            num_layers=args.num_layers,
            learning_rate=args.learning_rate,
            batch_size=args.batch_size,
            dropout=args.dropout,
            epoch=args.epoch,
            patience=args.patience,
            print_every=args.print_every,
            save_path=args.save_path,
        )


if __name__ == "__main__":
    entry_point()
